# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2024 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""The BIM Axis-related commands"""


import os
import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate
PARAMS = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/BIM")


class Arch_Axis:

    "the Arch Axis command definition"

    def GetResources(self):

        return {'Pixmap'  : 'Arch_Axis',
                'MenuText': QT_TRANSLATE_NOOP("Arch_Axis","Axis"),
                'Accel': "A, X",
                'ToolTip': QT_TRANSLATE_NOOP("Arch_Axis","Creates a set of axes")}

    def Activated(self):

        FreeCAD.ActiveDocument.openTransaction(translate("Arch","Create Axis"))
        FreeCADGui.addModule("Arch")

        FreeCADGui.doCommand("Arch.makeAxis()")
        FreeCAD.ActiveDocument.commitTransaction()

    def IsActive(self):

        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v



class Arch_AxisSystem:

    "the Arch Axis System command definition"

    def GetResources(self):

        return {'Pixmap'  : 'Arch_Axis_System',
                'MenuText': QT_TRANSLATE_NOOP("Arch_AxisSystem","Axis System"),
                'Accel': "X, S",
                'ToolTip': QT_TRANSLATE_NOOP("Arch_AxisSystem","Creates an axis system from a set of axes")}

    def Activated(self):

        import Draft
        if FreeCADGui.Selection.getSelection():
            s = "["
            for o in FreeCADGui.Selection.getSelection():
                if Draft.getType(o) != "Axis":
                    FreeCAD.Console.PrintError(translate("Arch","Only axes must be selected")+"\n")
                    return
                s += "FreeCAD.ActiveDocument."+o.Name+","
            s += "]"
            FreeCAD.ActiveDocument.openTransaction(translate("Arch","Create Axis System"))
            FreeCADGui.addModule("Arch")
            FreeCADGui.doCommand("Arch.makeAxisSystem("+s+")")
            FreeCAD.ActiveDocument.commitTransaction()
        else:
            FreeCAD.Console.PrintError(translate("Arch","Please select at least one axis")+"\n")

    def IsActive(self):

        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v


class Arch_Grid:

    "the Arch Grid command definition"

    def GetResources(self):

        return {'Pixmap'  : 'Arch_Grid',
                'MenuText': QT_TRANSLATE_NOOP("Arch_Grid","Grid"),
                'Accel': "A, X",
                'ToolTip': QT_TRANSLATE_NOOP("Arch_Grid","Creates a customizable grid object")}

    def Activated(self):

        FreeCAD.ActiveDocument.openTransaction(translate("Arch","Create Grid"))
        FreeCADGui.addModule("Arch")

        FreeCADGui.doCommand("Arch.makeGrid()")
        FreeCAD.ActiveDocument.commitTransaction()

    def IsActive(self):

        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v


class Arch_AxisTools:

    """The Axis tools group command"""

    def GetCommands(self):
        return tuple(['Arch_Axis','Arch_AxisSystem','Arch_Grid'])

    def GetResources(self):
        return { 'MenuText': QT_TRANSLATE_NOOP("Arch_AxisTools",'Axis tools'),
                 'ToolTip': QT_TRANSLATE_NOOP("Arch_AxisTools",'Axis tools')
               }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v



FreeCADGui.addCommand('Arch_Axis', Arch_Axis())
FreeCADGui.addCommand('Arch_AxisSystem', Arch_AxisSystem())
FreeCADGui.addCommand('Arch_Grid', Arch_Grid())
FreeCADGui.addCommand('Arch_AxisTools', Arch_AxisTools())



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022-2024 The FreeCAD Project Association AISBL         *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

""" Provides a class for showing the list view and detail view at the same time. """

import base64

from addonmanager_freecad_interface import Preferences

from Addon import Addon
from Widgets.addonmanager_widget_package_details_view import PackageDetailsView
from Widgets.addonmanager_widget_view_selector import AddonManagerDisplayStyle
from addonmanager_package_details_controller import PackageDetailsController
from package_list import PackageList

# Get whatever version of PySide we can
try:
    import PySide  # Use the FreeCAD wrapper
except ImportError:
    try:
        import PySide6 as PySide  # Outside FreeCAD, try Qt6 first
    except ImportError:
        # Fall back to Qt5 (if this fails, Python will kill this module's import)
        import PySide2 as PySide

from PySide import QtCore, QtWidgets


class CompositeView(QtWidgets.QWidget):
    """A widget that displays the Addon Manager's top bar, the list of Addons, and the detail
    view. Depending on the view mode selected, these may all be displayed at once, or selecting
    an addon in the list may case the list to hide and the detail view to show."""

    install = QtCore.Signal(Addon)
    uninstall = QtCore.Signal(Addon)
    update = QtCore.Signal(Addon)
    execute = QtCore.Signal(Addon)
    update_status = QtCore.Signal(Addon)
    check_for_update = QtCore.Signal(Addon)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.package_details = PackageDetailsView(self)
        self.package_details_controller = PackageDetailsController(self.package_details)
        self.package_list = PackageList(self)
        prefs = Preferences()
        self.display_style = prefs.get("ViewStyle")
        self.main_layout = QtWidgets.QHBoxLayout(self)
        self.splitter = QtWidgets.QSplitter(self)
        self.splitter.addWidget(self.package_list)
        self.splitter.addWidget(self.package_details)
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.splitter.setContentsMargins(0, 0, 0, 0)
        self.splitter.setSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding
        )
        self.main_layout.addWidget(self.splitter)
        self.layout().setContentsMargins(0, 0, 0, 0)
        self._setup_ui()
        self._setup_connections()
        self._restore_splitter_state()
        self.scroll_position = 0

    def _save_splitter_state(self):
        """Write the splitter state into an Addon manager preference, CompositeSplitterState"""
        prefs = Preferences()
        state = self.splitter.saveState()
        encoded = base64.b64encode(state).decode("ASCII")
        prefs.set("CompositeSplitterState", encoded)

    def _restore_splitter_state(self):
        """Restore the splitter state from CompositeSplitterState"""
        prefs = Preferences()
        encoded = prefs.get("CompositeSplitterState")
        if encoded:
            state = base64.b64decode(encoded)
            self.splitter.restoreState(state)

    def setModel(self, model):
        self.package_list.setModel(model)

    def set_display_style(self, style: AddonManagerDisplayStyle):
        self.display_style = style
        self._setup_ui()

    def _setup_ui(self):
        if self.display_style == AddonManagerDisplayStyle.EXPANDED:
            self._setup_expanded_ui()
        elif self.display_style == AddonManagerDisplayStyle.COMPACT:
            self._setup_compact_ui()
        elif self.display_style == AddonManagerDisplayStyle.COMPOSITE:
            self._setup_composite_ui()
        else:
            raise RuntimeError("Invalid display style")
        self.package_list.set_view_style(self.display_style)

    def _setup_expanded_ui(self):
        self.package_list.show()
        self.package_details.hide()
        self.package_details.button_bar.set_show_back_button(True)

    def _setup_compact_ui(self):
        self.package_list.show()
        self.package_details.hide()
        self.package_details.button_bar.set_show_back_button(True)

    def _setup_composite_ui(self):
        self.package_list.show()
        self.package_details.show()
        self.package_details.button_bar.set_show_back_button(False)

    def _setup_connections(self):
        self.package_list.itemSelected.connect(self.addon_selected)
        self.package_details_controller.back.connect(self._back_button_clicked)
        self.package_details_controller.install.connect(self.install)
        self.package_details_controller.uninstall.connect(self.uninstall)
        self.package_details_controller.update.connect(self.update)
        self.package_details_controller.execute.connect(self.execute)
        self.package_details_controller.update_status.connect(self.update_status)
        self.package_list.ui.view_bar.view_changed.connect(self.set_display_style)
        self.splitter.splitterMoved.connect(self._splitter_moved)

    def addon_selected(self, addon):
        """Depending on the display_style, show addon details (possibly hiding the package_list
        widget in the process."""
        self.package_details_controller.show_repo(addon)
        if self.display_style != AddonManagerDisplayStyle.COMPOSITE:
            self.scroll_position = (
                self.package_list.ui.listPackages.verticalScrollBar().sliderPosition()
            )
            print(f"Saved slider position at {self.scroll_position}")
            self.package_list.hide()
            self.package_details.show()
            self.package_details.button_bar.set_show_back_button(True)

    def _back_button_clicked(self):
        if self.display_style != AddonManagerDisplayStyle.COMPOSITE:
            print(f"Set slider position to {self.scroll_position}")
            self.package_list.show()
            self.package_details.hide()
            # The following must be done *after* a cycle through the event loop
            QtCore.QTimer.singleShot(
                0,
                lambda: self.package_list.ui.listPackages.verticalScrollBar().setSliderPosition(
                    self.scroll_position
                ),
            )

    def _splitter_moved(self, _1: int, _2: int) -> None:
        self._save_splitter_state()



# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2024 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""BIM window command"""


import os
import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate
PARAMS = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/BIM")
ALLOWEDHOSTS = ["Wall","Structure","Roof"]


class Arch_Window:

    "the Arch Window command definition"

    def __init__(self):

        self.doormode = False

    def GetResources(self):

        return {'Pixmap'  : 'Arch_Window',
                'MenuText': QT_TRANSLATE_NOOP("Arch_Window","Window"),
                'Accel': "W, N",
                'ToolTip': QT_TRANSLATE_NOOP("Arch_Window","Creates a window object from a selected object (wire, rectangle or sketch)")}

    def IsActive(self):

        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):

        from draftutils import params
        import Draft
        import WorkingPlane
        import draftguitools.gui_trackers as DraftTrackers
        self.sel = FreeCADGui.Selection.getSelection()
        self.W1 = params.get_param_arch("WindowW1")  # thickness of the fixed frame
        if self.doormode:
            self.Width = params.get_param_arch("DoorWidth")
            self.Height = params.get_param_arch("DoorHeight")
        else:
            self.Width = params.get_param_arch("WindowWidth")
            self.Height = params.get_param_arch("WindowHeight")
        self.RemoveExternal = params.get_param_arch("archRemoveExternal")
        self.Preset = 0
        self.LibraryPreset = 0
        self.Sill = 0
        self.Include = True
        self.baseFace = None
        self.wparams = ["Width","Height","H1","H2","H3","W1","W2","O1","O2"]
        self.wp = None

        # autobuild mode
        if FreeCADGui.Selection.getSelectionEx():
            FreeCADGui.draftToolBar.offUi()
            obj = self.sel[0]
            if hasattr(obj,'Shape'):
                if obj.Shape.Wires and (not obj.Shape.Solids) and (not obj.Shape.Shells):
                    FreeCADGui.Control.closeDialog()
                    host = None
                    if hasattr(obj,"AttachmentSupport"):
                        if obj.AttachmentSupport:
                            if isinstance(obj.AttachmentSupport,tuple):
                                host = obj.AttachmentSupport[0]
                            elif isinstance(obj.AttachmentSupport,list):
                                host = obj.AttachmentSupport[0][0]
                            else:
                                host = obj.AttachmentSupport
                            obj.AttachmentSupport = None # remove
                    elif Draft.isClone(obj,"Window"):
                        if obj.Objects[0].Inlist:
                            host = obj.Objects[0].Inlist[0]

                    FreeCAD.ActiveDocument.openTransaction(translate("Arch","Create Window"))
                    FreeCADGui.addModule("Arch")
                    FreeCADGui.doCommand("win = Arch.makeWindow(FreeCAD.ActiveDocument."+obj.Name+")")
                    if host and self.Include:
                        FreeCADGui.doCommand("win.Hosts = [FreeCAD.ActiveDocument."+host.Name+"]")
                        siblings = host.Proxy.getSiblings(host)
                        sibs = [host]
                        for sibling in siblings:
                            if not sibling in sibs:
                                sibs.append(sibling)
                                FreeCADGui.doCommand("win.Hosts = win.Hosts+[FreeCAD.ActiveDocument."+sibling.Name+"]")
                    FreeCAD.ActiveDocument.commitTransaction()
                    FreeCAD.ActiveDocument.recompute()
                    return

                # Try to detect an object to use as a window type - TODO we must make this safer

                elif obj.Shape.Solids and (Draft.getType(obj) not in ["Wall","Structure","Roof"]):
                    # we consider the selected object as a type
                    FreeCAD.ActiveDocument.openTransaction(translate("Arch","Create Window"))
                    FreeCADGui.addModule("Arch")
                    FreeCADGui.doCommand("Arch.makeWindow(FreeCAD.ActiveDocument."+obj.Name+")")
                    FreeCAD.ActiveDocument.commitTransaction()
                    FreeCAD.ActiveDocument.recompute()
                    return

        # interactive mode
        self.wp = WorkingPlane.get_working_plane()

        self.tracker = DraftTrackers.boxTracker()
        self.tracker.length(self.Width)
        self.tracker.width(self.W1)
        self.tracker.height(self.Height)
        self.tracker.on()
        FreeCAD.Console.PrintMessage(translate("Arch","Choose a face on an existing object or select a preset")+"\n")
        FreeCADGui.Snapper.getPoint(callback=self.getPoint,movecallback=self.update,extradlg=self.taskbox())
        #FreeCADGui.Snapper.setSelectMode(True)

    def has_width_and_height_constraint(self, sketch):
        width_found = False
        height_found = False

        for constr in sketch.Constraints:
            if constr.Name == "Width":
                width_found = True
            elif constr.Name == "Height":
                height_found = True
            elif width_found and height_found:
                break

        return (width_found and height_found)

    def getPoint(self,point=None,obj=None):

        "this function is called by the snapper when it has a 3D point"

        import Draft
        from draftutils import gui_utils
        from draftutils.messages import _wrn
        from ArchWindowPresets import WindowPresets
        self.tracker.off()
        if point is None:
            return
        # if something was selected, override the underlying object
        if self.sel:
            obj = self.sel[0]
        point = point.add(FreeCAD.Vector(0,0,self.Sill))
        FreeCAD.ActiveDocument.openTransaction(translate("Arch","Create Window"))

        FreeCADGui.doCommand("import math, FreeCAD, Arch, DraftGeomUtils, WorkingPlane")
        FreeCADGui.doCommand("wp = WorkingPlane.get_working_plane()")

        if self.baseFace is not None:
            FreeCADGui.doCommand("face = FreeCAD.ActiveDocument." + self.baseFace[0].Name + ".Shape.Faces[" + str(self.baseFace[1]) + "]")
            FreeCADGui.doCommand("pl = DraftGeomUtils.placement_from_face(face, vec_z = wp.axis)")
        else:
            FreeCADGui.doCommand("pl = FreeCAD.Placement()")
            FreeCADGui.doCommand("pl.Rotation = FreeCAD.Rotation(wp.u, wp.axis, -wp.v, 'XZY')")

        FreeCADGui.doCommand("pl.Base = FreeCAD.Vector(" + str(point.x) + ", " + str(point.y) + ", " + str(point.z) + ")")

        if self.Preset >= len(WindowPresets):
            # library object
            col = FreeCAD.ActiveDocument.Objects
            path = self.librarypresets[self.Preset - len(WindowPresets)][1]
            FreeCADGui.doCommand("FreeCADGui.ActiveDocument.mergeProject('" + path + "')")
            # find the latest added window
            nol = FreeCAD.ActiveDocument.Objects
            for o in nol[len(col):]:
                if Draft.getType(o) == "Window":
                    if Draft.getType(o.Base) != "Sketcher::SketchObject":
                        _wrn(translate("Arch", "Window not based on sketch. Window not aligned or resized."))
                        self.Include = False
                        break
                    FreeCADGui.doCommand("win = FreeCAD.ActiveDocument.getObject('" + o.Name + "')")
                    FreeCADGui.doCommand("win.Base.Placement = pl")
                    FreeCADGui.doCommand("win.Normal = pl.Rotation.multVec(FreeCAD.Vector(0, 0, -1))")
                    FreeCADGui.doCommand("win.Width = " + str(self.Width))
                    FreeCADGui.doCommand("win.Height = " + str(self.Height))
                    FreeCADGui.doCommand("win.Base.recompute()")
                    if not self.has_width_and_height_constraint(o.Base):
                        _wrn(translate("Arch", "No Width and/or Height constraint in window sketch. Window not resized."))
                    break
            else:
                _wrn(translate("Arch", "No window found. Cannot continue."))
                self.Include = False

        else:
            # preset
            wp = ""
            for p in self.wparams:
                wp += p.lower() + "=" + str(getattr(self,p)) + ", "
            FreeCADGui.doCommand("win = Arch.makeWindowPreset('" + WindowPresets[self.Preset] + "', " + wp + "placement=pl)")

        if self.Include:
            host = None
            if self.baseFace is not None:
                host = self.baseFace[0]
            elif obj:
                host = obj
            if Draft.getType(host) in ALLOWEDHOSTS:
                FreeCADGui.doCommand("win.Hosts = [FreeCAD.ActiveDocument." + host.Name + "]")
                siblings = host.Proxy.getSiblings(host)
                for sibling in siblings:
                    FreeCADGui.doCommand("win.Hosts = win.Hosts + [FreeCAD.ActiveDocument." + sibling.Name + "]")

        FreeCAD.ActiveDocument.commitTransaction()
        FreeCAD.ActiveDocument.recompute()
        # gui_utils.end_all_events()  # Causes a crash on Linux.
        self.tracker.finalize()
        return

    def update(self,point,info):

        "this function is called by the Snapper when the mouse is moved"

        delta = FreeCAD.Vector(self.Width/2,self.W1/2,self.Height/2)
        delta = delta.add(FreeCAD.Vector(0,0,self.Sill))

        if self.baseFace is None:
            rot = FreeCAD.Rotation(self.wp.u,self.wp.v,-self.wp.axis,"XZY")
            self.tracker.setRotation(rot)
        if info:
            if "Face" in info['Component']:
                import DraftGeomUtils
                o = FreeCAD.ActiveDocument.getObject(info['Object'])
                self.baseFace = [o,int(info['Component'][4:])-1]
                #print("switching to ",o.Label," face ",self.baseFace[1])
                f = o.Shape.Faces[self.baseFace[1]]
                p = DraftGeomUtils.placement_from_face(f,vec_z=self.wp.axis,rotated=True)
                rot = p.Rotation
                self.tracker.setRotation(rot)
        r = self.tracker.trans.rotation.getValue().getValue()
        if r != (0,0,0,1):
            delta = FreeCAD.Rotation(r[0],r[1],r[2],r[3]).multVec(FreeCAD.Vector(delta.x,-delta.y,-delta.z))
        self.tracker.pos(point.add(delta))

    def taskbox(self):

        "sets up a taskbox widget"

        from draftutils import params
        from PySide import QtCore, QtGui, QtSvgWidgets
        from ArchWindowPresets import WindowPresets
        w = QtGui.QWidget()
        ui = FreeCADGui.UiLoader()
        w.setWindowTitle(translate("Arch","Window options"))
        grid = QtGui.QGridLayout(w)

        # include box
        include = QtGui.QCheckBox(translate("Arch","Auto include in host object"))
        include.setChecked(True)
        grid.addWidget(include,0,0,1,2)
        include.stateChanged.connect(self.setInclude)

        # sill height
        labels = QtGui.QLabel(translate("Arch","Sill height"))
        values = ui.createWidget("Gui::InputField")
        grid.addWidget(labels,1,0,1,1)
        grid.addWidget(values,1,1,1,1)
        values.valueChanged.connect(self.setSill)

        # check for Parts library and Arch presets

        # because of the use of FreeCADGui.doCommand() backslashes in the
        # paths in librarypresets need to be double escaped "\\\\", so let's
        # use forward slashes instead...
        self.librarypresets = []
        librarypath = FreeCAD.ParamGet("User parameter:Plugins/parts_library").GetString("destination", "")
        # librarypath should have only forward slashes already, but let's use replace() anyway just to be sure:
        librarypath = librarypath.replace("\\", "/") + "/Architectural Parts"
        presetdir = FreeCAD.getUserAppDataDir().replace("\\", "/") + "/Arch"
        for path in [librarypath, presetdir]:
            if os.path.isdir(path):
                for wtype in ["Windows", "Doors"]:
                    wdir = path + "/" + wtype
                    if os.path.isdir(wdir):
                        for subtype in os.listdir(wdir):
                            subdir = wdir + "/" + subtype
                            if os.path.isdir(subdir):
                                for subfile in os.listdir(subdir):
                                    if (os.path.isfile(subdir + "/" + subfile)
                                            and subfile.lower().endswith(".fcstd")):
                                        self.librarypresets.append([wtype + " - " + subtype + " - " + subfile[:-6],
                                                                    subdir + "/" + subfile])

        # presets box
        labelp = QtGui.QLabel(translate("Arch","Preset"))
        valuep = QtGui.QComboBox()
        valuep.setMinimumContentsLength(6)
        valuep.setSizeAdjustPolicy(QtGui.QComboBox.AdjustToContents)
        valuep.addItems(WindowPresets)
        valuep.setCurrentIndex(self.Preset)
        grid.addWidget(labelp,2,0,1,1)
        grid.addWidget(valuep,2,1,1,1)
        valuep.currentIndexChanged.connect(self.setPreset)
        for it in self.librarypresets:
            valuep.addItem(it[0])

        # image display
        self.pic = QtGui.QLabel()
        grid.addWidget(self.pic,3,0,1,2)
        self.pic.setFixedHeight(128)
        self.pic.hide()

        # SVG display
        self.im = QtSvgWidgets.QSvgWidget(":/ui/ParametersWindowFixed.svg")
        self.im.setMaximumWidth(200)
        self.im.setMinimumHeight(120)
        grid.addWidget(self.im,4,0,1,2)
        #self.im.hide()

        # parameters
        i = 5
        for param in self.wparams:
            lab = QtGui.QLabel(translate("Arch",param))
            setattr(self,"val"+param,ui.createWidget("Gui::InputField"))
            wid = getattr(self,"val"+param)
            if param == "W1":
                wid.setText(FreeCAD.Units.Quantity(self.W1,FreeCAD.Units.Length).UserString)
            elif param == "Width":
                wid.setText(FreeCAD.Units.Quantity(self.Width,FreeCAD.Units.Length).UserString)
            elif param == "Height":
                wid.setText(FreeCAD.Units.Quantity(self.Height,FreeCAD.Units.Length).UserString)
            else:
                n = params.get_param_arch("Window"+param)
                wid.setText(FreeCAD.Units.Quantity(n,FreeCAD.Units.Length).UserString)
                setattr(self,param,n)
            grid.addWidget(lab,i,0,1,1)
            grid.addWidget(wid,i,1,1,1)
            i += 1
            valueChanged = self.getValueChanged(param)
            FreeCAD.wid = wid
            QtCore.QObject.connect(getattr(self,"val"+param),QtCore.SIGNAL("valueChanged(double)"), valueChanged)

        # restore saved states
        if self.doormode:
            i = params.get_param_arch("DoorPreset")
            d = params.get_param_arch("DoorSill")
        else:
            i = params.get_param_arch("WindowPreset")
            d = params.get_param_arch("WindowSill")
        if i < valuep.count():
            valuep.setCurrentIndex(i)
        values.setText(FreeCAD.Units.Quantity(d,FreeCAD.Units.Length).UserString)

        return w

    def getValueChanged(self,p):

        return lambda d : self.setParams(p, d)

    def setSill(self,d):

        from draftutils import params
        self.Sill = d
        if self.doormode:
            params.set_param_arch("DoorSill",d)
        else:
            params.set_param_arch("WindowSill",d)

    def setInclude(self,i):

        self.Include = bool(i)

    def setParams(self,param,d):

        from draftutils import params
        setattr(self,param,d)
        self.tracker.length(self.Width)
        self.tracker.height(self.Height)
        self.tracker.width(self.W1)
        prefix = "Door" if self.doormode and param in ("Width","Height") else "Window"
        params.set_param_arch(prefix+param,d)

    def setPreset(self,i):

        from PySide import QtGui
        from draftutils import params
        from ArchWindowPresets import WindowPresets
        self.Preset = i
        if self.doormode:
            params.set_param_arch("DoorPreset",i)
        else:
            params.set_param_arch("WindowPreset",i)
        if i >= 0:
            FreeCADGui.Snapper.setSelectMode(False)
            self.tracker.length(self.Width)
            self.tracker.height(self.Height)
            self.tracker.width(self.W1)
            self.tracker.on()
            self.pic.hide()
            self.im.show()
            if i == 0:
                self.im.load(":/ui/ParametersWindowFixed.svg")
            elif i in [1,8]:
                self.im.load(":/ui/ParametersWindowSimple.svg")
            elif i in [2,4,7]:
                self.im.load(":/ui/ParametersWindowDouble.svg")
            elif i == 3:
                self.im.load(":/ui/ParametersWindowStash.svg")
            elif i == 5:
                self.im.load(":/ui/ParametersDoorSimple.svg")
            elif i == 6:
                self.im.load(":/ui/ParametersDoorGlass.svg")
            elif i == 9:
                self.im.load(":/ui/ParametersOpening.svg")
            else:
                # From Library
                self.im.hide()
                path = self.librarypresets[i-len(WindowPresets)][1]
                if path.lower().endswith(".fcstd"):
                    try:
                        import tempfile
                        import zipfile
                    except Exception:
                        pass
                    else:
                        zfile = zipfile.ZipFile(path)
                        files = zfile.namelist()
                        # check for meta-file if it's really a FreeCAD document
                        if files[0] == "Document.xml":
                            image="thumbnails/Thumbnail.png"
                            if image in files:
                                image = zfile.read(image)
                                thumbfile = tempfile.mkstemp(suffix='.png')[1]
                                thumb = open(thumbfile,"wb")
                                thumb.write(image)
                                thumb.close()
                                im = QtGui.QPixmap(thumbfile)
                                self.pic.setPixmap(im)
                                self.pic.show()
            #for param in self.wparams:
            #    getattr(self,"val"+param).setEnabled(True)
        else:
            FreeCADGui.Snapper.setSelectMode(True)
            self.tracker.off()
            self.im.hide()
            for param in self.wparams:
                getattr(self,"val"+param).setEnabled(False)


FreeCADGui.addCommand('Arch_Window', Arch_Window())



# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2021 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************


"""This module contains FreeCAD commands for the BIM workbench"""

import os
import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate

class BIM_Reorder:
    def GetResources(self):
        return {
            "Pixmap": "BIM_Reorder",
            "MenuText": QT_TRANSLATE_NOOP("BIM_Reorder", "Reorder children"),
            # 'Accel': "R, D",
            "ToolTip": QT_TRANSLATE_NOOP(
                "BIM_Reorder", "Reorder children of selected object"
            ),
        }

    def Activated(self):
        if len(FreeCADGui.Selection.getSelection()) == 1:
            obj = FreeCADGui.Selection.getSelection()[0]
            if hasattr(obj, "Group"):
                if obj.getTypeIdOfProperty("Group") == "App::PropertyLinkList":
                    FreeCADGui.Control.showDialog(BIM_Reorder_TaskPanel(obj))
                    return
        FreeCAD.Console.PrintError(
            translate("BIM", "You must choose a group object before using this command")
            + "\n"
        )


class BIM_Reorder_TaskPanel:

    def __init__(self, obj):
        from PySide import QtCore, QtGui

        self.obj = obj
        self.form = FreeCADGui.PySideUic.loadUi(":/ui/dialogReorder.ui")
        self.form.setWindowIcon(QtGui.QIcon(":/icons/BIM_Reorder.svg"))
        self.form.pushButton.clicked.connect(self.form.listWidget.sortItems)
        for child in self.obj.Group:
            i = QtGui.QListWidgetItem(child.Label)
            i.setIcon(child.ViewObject.Icon)
            i.setToolTip(child.Name)
            self.form.listWidget.addItem(i)

    def accept(self):
        names = [
            self.form.listWidget.item(i).toolTip()
            for i in range(self.form.listWidget.count())
        ]
        group = [FreeCAD.ActiveDocument.getObject(n) for n in names]
        FreeCAD.ActiveDocument.openTransaction("Reorder children")
        self.obj.Group = group
        FreeCAD.ActiveDocument.commitTransaction()
        FreeCADGui.doCommand("FreeCAD.ActiveDocument.recompute()")
        return self.reject()

    def reject(self):
        FreeCADGui.Control.closeDialog()
        FreeCAD.ActiveDocument.recompute()
        return True


FreeCADGui.addCommand("BIM_Reorder", BIM_Reorder())



#***************************************************************************
#*   Copyright (c) 2018 Yorik van Havre <yorik@uncreated.net>              *
#*                                                                         *
#*   This program is free software; you can redistribute it and/or modify  *
#*   it under the terms of the GNU Lesser General Public License (LGPL)    *
#*   as published by the Free Software Foundation; either version 2 of     *
#*   the License, or (at your option) any later version.                   *
#*   for detail see the LICENCE text file.                                 *
#*                                                                         *
#*   This program is distributed in the hope that it will be useful,       *
#*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
#*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
#*   GNU Library General Public License for more details.                  *
#*                                                                         *
#*   You should have received a copy of the GNU Library General Public     *
#*   License along with this program; if not, write to the Free Software   *
#*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
#*   USA                                                                   *
#*                                                                         *
#***************************************************************************

__title__  = "FreeCAD Arch External Reference"
__author__ = "Yorik van Havre"
__url__    = "https://www.freecad.org"


import FreeCAD
import os
import zipfile
import re
from draftutils import params

if FreeCAD.GuiUp:
    import FreeCADGui
    from PySide import QtCore, QtGui
    from draftutils.translate import translate
    from PySide.QtCore import QT_TRANSLATE_NOOP
else:
    # \cond
    def translate(ctxt,txt):
        return txt
    def QT_TRANSLATE_NOOP(ctxt,txt):
        return txt
    # \endcond

## @package ArchReference
#  \ingroup ARCH
#  \brief The Reference object and tools
#
#  This module provides tools to build Reference objects.
#  References can take a shape from a Part-based object in
#  another file.


class ArchReference:

    """The Arch Reference object"""

    def __init__(self, obj):

        obj.Proxy = self
        ArchReference.setProperties(self, obj)
        self.Type = "Reference"
        self.reload = True


    def setProperties(self, obj):

        pl = obj.PropertiesList
        if not "File" in pl:
            t = QT_TRANSLATE_NOOP("App::Property","The base file this component is built upon")
            obj.addProperty("App::PropertyFile","File","Reference",t)
        if not "Part" in pl:
            t = QT_TRANSLATE_NOOP("App::Property","The part to use from the base file")
            obj.addProperty("App::PropertyString","Part","Reference",t)
        if not "ReferenceMode" in pl:
            t = QT_TRANSLATE_NOOP("App::Property","The way the referenced objects are included in the current document. 'Normal' includes the shape, 'Transient' discards the shape when the object is switched off (smaller filesize), 'Lightweight' does not import the shape but only the OpenInventor representation")
            obj.addProperty("App::PropertyEnumeration","ReferenceMode","Reference",t)
            obj.ReferenceMode = ["Normal","Transient","Lightweight"]
            if "TransientReference" in pl:
                if obj.TransientReference:
                    obj.ReferenceMode = "Transient"
                obj.removeProperty("TransientReference")
                t = translate("Arch", "TransientReference property to ReferenceMode")
                FreeCAD.Console.PrintMessage(translate("Arch","Upgrading")+" "+obj.Label+" "+t+"\n")
        if not "FuseArch" in pl:
            t = QT_TRANSLATE_NOOP("App::Property","Fuse objects of same material")
            obj.addProperty("App::PropertyBool","FuseArch", "Reference", t)
        self.Type = "Reference"


    def onDocumentRestored(self, obj):

        ArchReference.setProperties(self, obj)
        self.reload = False
        if obj.ReferenceMode == "Lightweight":
            if obj.ViewObject and obj.ViewObject.Proxy:
                obj.ViewObject.Proxy.loadInventor(obj)


    def dumps(self):

        return None


    def loads(self, state):

        return None


    def onChanged(self, obj, prop):

        if prop in ["File","Part"]:
            self.reload = True
        elif prop == "ReferenceMode":
            if obj.ReferenceMode == "Normal":
                if obj.ViewObject and obj.ViewObject.Proxy:
                    obj.ViewObject.Proxy.unloadInventor(obj)
                if (not obj.Shape) or obj.Shape.isNull():
                    self.reload = True
                    obj.touch()
            elif obj.ReferenceMode == "Transient":
                if obj.ViewObject and obj.ViewObject.Proxy:
                    obj.ViewObject.Proxy.unloadInventor(obj)
                self.reload = False
            elif obj.ReferenceMode == "Lightweight":
                self.reload = False
                import Part
                pl = obj.Placement
                obj.Shape = Part.Shape()
                obj.Placement = pl
                if obj.ViewObject and obj.ViewObject.Proxy:
                    obj.ViewObject.Proxy.loadInventor(obj)


    def execute(self, obj):

        import Part
        pl = obj.Placement
        filename = self.getFile(obj)
        if filename and self.reload and obj.ReferenceMode in ["Normal","Transient"]:
            self.parts = self.getPartsList(obj)
            if self.parts:
                if filename.lower().endswith(".fcstd"):
                    zdoc = zipfile.ZipFile(filename)
                    if zdoc:
                        if obj.Part:
                            if obj.Part in self.parts:
                                if self.parts[obj.Part][1] in zdoc.namelist():
                                    f = zdoc.open(self.parts[obj.Part][1])
                                    shapedata = f.read()
                                    f.close()
                                    shapedata = shapedata.decode("utf8")
                                    shape = self.cleanShape(shapedata,obj,self.parts[obj.Part][2])
                                    obj.Shape = shape
                                    if not pl.isIdentity():
                                        obj.Placement = pl
                                else:
                                    t = translate("Arch","Part not found in file")
                                    FreeCAD.Console.PrintError(t+"\n")
                        else:
                            shapes = []
                            for part in self.parts.values():
                                f = zdoc.open(part[1])
                                shapedata = f.read()
                                f.close()
                                shapedata = shapedata.decode("utf8")
                                shape = self.cleanShape(shapedata,obj)
                                shapes.append(shape)
                            if shapes:
                                obj.Shape = Part.makeCompound(shapes)
                elif filename.lower().endswith(".ifc"):
                    ifcfile = self.getIfcFile(filename)
                    if not ifcfile:
                        return
                    try:
                        from nativeifc import ifc_tools
                        from nativeifc import ifc_generator
                    except:
                        t = translate("Arch","NativeIFC not available - unable to process IFC files")
                        FreeCAD.Console.PrintError(t+"\n")
                        return
                    elements = self.getIFCElements(obj, ifcfile)
                    shape, colors = ifc_generator.generate_shape(ifcfile, elements, cached=True)
                    if shape:
                        placement = shape.Placement
                        obj.Shape = shape
                        obj.Placement = placement
                        if colors:
                            ifc_tools.set_colors(obj, colors)
            elif filename.lower().endswith(".dxf"):
                # create a special parameter set to control the DXF importer
                loc = "User parameter:BaseApp/Preferences/Mod/Arch"
                hGrp = FreeCAD.ParamGet(loc).GetGroup("RefDxfImport")
                hGrp.SetBool("dxfUseDraftVisGroups", False)
                hGrp.SetBool("dxfGetOriginalColors", False)
                hGrp.SetBool("groupLayers", True)
                hGrp.SetFloat("dxfScaling", 1.0)
                hGrp.SetBool("dxftext", False)
                hGrp.SetBool("dxfImportPoints", True)
                hGrp.SetBool("dxflayout", False)
                hGrp.SetBool("dxfstarblocks", False)
                doc = obj.Document
                oldobjs = list(doc.Objects)
                import Import
                Import.readDXF(filename, doc.Name, True, loc + "/RefDxfImport")
                newobjs = [o for o in doc.Objects if o not in oldobjs]
                shapes = [o.Shape for o in newobjs if o.isDerivedFrom("Part::Feature")]
                if len(shapes) == 1:
                    obj.Shape = shapes[0]
                elif len(shapes) > 1:
                    obj.Shape = Part.makeCompound(shapes)
                names = [o.Name for o in newobjs]
                for n in names:
                    doc.removeObject(n)
            self.reload = False


    def getIFCElements(self, obj, ifcfile):

        """returns IFC elements for this object"""

        try:
            from nativeifc import ifc_generator
        except:
            t = translate("Arch","NativeIFC not available - unable to process IFC files")
            FreeCAD.Console.PrintError(t+"\n")
            return
        if obj.Part:
            element = ifcfile[int(obj.Part)]
        else:
            element = ifcfile.by_type("IfcProject")[0]
        elements = ifc_generator.get_decomposed_elements(element)
        elements = ifc_generator.filter_types(elements)
        return elements


    def cleanShape(self, shapedata, obj, materials=None):

        """cleans the imported shape"""

        import Part
        shape = Part.Shape()
        shape.importBrepFromString(shapedata)
        if obj.FuseArch and materials:
            # separate lone edges
            shapes = []
            for edge in shape.Edges:
                found = False
                for solid in shape.Solids:
                    for soledge in solid.Edges:
                        if edge.hashCode() == soledge.hashCode():
                            found = True
                            break
                    if found:
                        break
                if found:
                    break
            else:
                shapes.append(edge)
            #print("solids:",len(shape.Solids),"mattable:",materials)
            for key,solindexes in materials.items():
                if key == "Undefined":
                    # do not join objects with no defined material
                    for solindex in [int(i) for i in solindexes.split(",")]:
                        shapes.append(shape.Solids[solindex])
                else:
                    fusion = None
                    for solindex in [int(i) for i in solindexes.split(",")]:
                        if not fusion:
                            fusion = shape.Solids[solindex]
                        else:
                            fusion = fusion.fuse(shape.Solids[solindex])
                    if fusion:
                        shapes.append(fusion)
            shape = Part.makeCompound(shapes)
            try:
                shape = shape.removeSplitter()
            except Exception:
                t = translate("Arch","Error removing splitter")
                FreeCAD.Console.PrintError(obj.Label+": "+t+"\n")
        return shape


    def exists(self,filepath):

        """case-insensitive version of os.path.exists. Returns the actual file path or None"""

        if os.path.exists(filepath):
            return filepath
        # check for uppercase/lowercase extensions
        p, e = os.path.splitext(filepath)
        if os.path.exists(p + e.lower()):
            return p + e.lower()
        if os.path.exists(p + e.upper()):
            return p + e.upper()
        return None


    def getFile(self,obj,filename=None):

        """gets a valid file, if possible"""

        if not filename:
            filename = obj.File
        if not filename:
            return None
        if not filename.lower().endswith(".fcstd"):
            if not filename.lower().endswith(".ifc"):
                if not filename.lower().endswith(".dxf"):
                    return None
        if not self.exists(filename):
            # search for the file in the current directory if not found
            basename = os.path.basename(filename)
            currentdir = os.path.dirname(obj.Document.FileName)
            altfile = os.path.join(currentdir,basename)
            if altfile == obj.Document.FileName:
                return None
            elif self.exists(altfile):
                return self.exists(altfile)
            else:
                # search for subpaths in current folder
                altfile = None
                subdirs = self.splitall(os.path.dirname(filename))
                for i in range(len(subdirs)):
                    subpath = [currentdir]+subdirs[-i:]+[basename]
                    altfile = os.path.join(*subpath)
                    if self.exists(altfile):
                        return self.exists(altfile)
                return None
        return self.exists(filename)


    def getPartsList(self, obj, filename=None):

        """returns a list of Part-based objects in a file"""

        filename = self.getFile(obj, filename)
        if not filename:
            return None
        if filename.lower().endswith(".fcstd"):
            return self.getPartsListFCSTD(obj, filename)
        elif filename.lower().endswith(".ifc"):
            return self.getPartsListIFC(obj, filename)
        elif filename.lower().endswith(".dxf"):
            return self.getPartsListDXF(obj, filename)


    def getPartsListDXF(self, obj, filename):

        """returns a list of Part-based objects in a DXF file"""

        # support layers
        #with open(filename) as f:
        #    txt = f.read()
        return {}


    def getPartsListIFC(self, obj, filename):

        """returns a list of Part-based objects in a IFC file"""

        ifcfile = self.getIfcFile(filename)
        if not ifcfile:
            return None
        structs = ifcfile.by_type("IfcSpatialElement")
        res = {}
        for s in structs:
            n = s.Name
            if not n:
                n = ""
            name = "#" + str(s.id()) + " " + n + "(" + s.is_a() + ")"
            res[str(s.id())] = [name, s, None]
        return res


    def getPartsListFCSTD(self, obj, filename):

        """returns a list of Part-based objects in a FCStd file"""

        parts = {}
        materials = {}
        zdoc = zipfile.ZipFile(filename)
        with zdoc.open("Document.xml") as docf:
            name = None
            label = None
            part = None
            materials = {}
            writemode = False
            for line in docf:
                line = line.decode("utf8")
                if "<Object name=" in line:
                    n = re.findall(r'name=\"(.*?)\"',line)
                    if n:
                        name = n[0]
                elif "<Property name=\"Label\"" in line:
                    writemode = True
                elif writemode and "<String value=" in line:
                    n = re.findall(r'value=\"(.*?)\"',line)
                    if n:
                        label = n[0]
                        writemode = False
                elif "<Property name=\"Shape\" type=\"Part::PropertyPartShape\"" in line:
                    writemode = True
                elif writemode and "<Part" in line and "file=" in line:
                    n = re.findall(r'file=\"(.*?)\"',line)
                    if n:
                        part = n[0]
                        writemode = False
                elif "<Property name=\"MaterialsTable\" type=\"App::PropertyMap\"" in line:
                    writemode = True
                elif writemode and "<Item key=" in line:
                    n = re.findall(r'key=\"(.*?)\"',line)
                    v = re.findall(r'value=\"(.*?)\"',line)
                    if n and v:
                        materials[n[0]] = v[0]
                elif writemode and "</Map>" in line:
                    writemode = False
                elif "</Object>" in line:
                    if name and label and part:
                        parts[name] = [label,part,materials]
                    name = None
                    label = None
                    part = None
                    materials = {}
                    writemode = False
        return parts


    def getIfcFile(self, filename):

        """Gets an IfcOpenShell object"""

        try:
            import ifcopenshell
        except:
            t = translate("Arch","NativeIFC not available - unable to process IFC files")
            FreeCAD.Console.PrintError(t+"\n")
            return None
        if not getattr(self, "ifcfile", None):
            self.ifcfile = ifcopenshell.open(filename)
        return self.ifcfile


    def getColors(self, obj):

        """returns the DiffuseColor of the referenced object"""

        filename = self.getFile(obj)
        if not filename:
            return None
        part = obj.Part
        if not obj.Part:
            return None
        colors = None
        if filename.lower().endswith(".fcstd"):
            zdoc = zipfile.ZipFile(filename)
            if not "GuiDocument.xml" in zdoc.namelist():
                return None
            colorfile = None
            with zdoc.open("GuiDocument.xml") as docf:
                writemode1 = False
                writemode2 = False
                for line in docf:
                    line = line.decode("utf8")
                    if ("<ViewProvider name=" in line) and (part in line):
                        writemode1 = True
                    elif writemode1 and ("<Property name=\"DiffuseColor\"" in line):
                        writemode1 = False
                        writemode2 = True
                    elif writemode2 and ("<ColorList file=" in line):
                        n = re.findall(r'file=\"(.*?)\"',line)
                        if n:
                            colorfile = n[0]
                            break
            if not colorfile:
                return None
            if not colorfile in zdoc.namelist():
                return None
            colors = []
            cf = zdoc.open(colorfile)
            buf = cf.read()
            cf.close()
            for i in range(1,int(len(buf)/4)):
                colors.append((buf[i*4+3]/255.0,buf[i*4+2]/255.0,buf[i*4+1]/255.0,buf[i*4]/255.0))
        return colors


    def splitall(self,path):

        """splits a path between its components"""

        allparts = []
        while 1:
            parts = os.path.split(path)
            if parts[0] == path:  # sentinel for absolute paths
                allparts.insert(0, parts[0])
                break
            elif parts[1] == path: # sentinel for relative paths
                allparts.insert(0, parts[1])
                break
            else:
                path = parts[0]
                allparts.insert(0, parts[1])
        return allparts


class ViewProviderArchReference:


    """A View Provider for the Arch Reference object"""


    def __init__(self,vobj):

        vobj.Proxy = self
        self.setProperties(vobj)


    def setProperties(self,vobj):

        pl = vobj.PropertiesList
        if not "TimeStamp" in pl:
            t = QT_TRANSLATE_NOOP("App::Property","The latest time stamp of the linked file")
            vobj.addProperty("App::PropertyFloat","TimeStamp","Reference",t)
            vobj.setEditorMode("TimeStamp",2)
        if not "UpdateColors" in pl:
            t = QT_TRANSLATE_NOOP("App::Property","If true, the colors from the linked file will be kept updated")
            vobj.addProperty("App::PropertyBool","UpdateColors","Reference",t)
            vobj.UpdateColors = True


    def getIcon(self):

        import Arch_rc
        return ":/icons/Arch_Reference.svg"


    def attach(self,vobj):

        self.Object = vobj.Object
        # Check for file change every minute
        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.checkChanges)
        s = params.get_param_arch("ReferenceCheckInterval")
        self.timer.start(1000*s)


    def dumps(self):

        return None


    def loads(self,state):

        return None


    def updateData(self,obj,prop):

        if (prop == "Shape"):
            if hasattr(obj.ViewObject,"UpdateColors") and obj.ViewObject.UpdateColors:
                if obj.Shape and not obj.Shape.isNull():
                    colors = obj.Proxy.getColors(obj)
                    if colors:
                        obj.ViewObject.DiffuseColor = colors
                    from DraftGui import todo
                    todo.delay(self.recolorize,obj.ViewObject)


    def recolorize(self,vobj):

        if hasattr(vobj,"DiffuseColor") and hasattr(vobj,"UpdateColors") and vobj.UpdateColors:
            vobj.DiffuseColor = vobj.DiffuseColor


    def checkChanges(self):

        "checks if the linked file has changed"

        if hasattr(self,"Object") and self.Object:
            try:
                f = self.Object.File
            except ReferenceError:
                f = None
                if hasattr(self,"timer"):
                    self.timer.stop()
                    del self.timer
            if f:
                filename = self.Object.Proxy.getFile(self.Object)
                if filename:
                    st_mtime = os.stat(filename).st_mtime
                    if hasattr(self.Object.ViewObject,"TimeStamp"):
                        if self.Object.ViewObject.TimeStamp:
                            if self.Object.ViewObject.TimeStamp != st_mtime:
                                self.Object.Proxy.reload = True
                                self.Object.touch()
                        self.Object.ViewObject.TimeStamp = st_mtime


    def onChanged(self,vobj,prop):

        if prop == "ShapeColor":
            # prevent ShapeColor from overriding DiffuseColor
            if hasattr(vobj,"DiffuseColor") and hasattr(vobj,"UpdateColors"):
                if vobj.DiffuseColor and vobj.UpdateColors:
                    vobj.DiffuseColor = vobj.DiffuseColor
        elif prop == "Visibility":
            if vobj.Visibility:
                if (not vobj.Object.Shape) or vobj.Object.Shape.isNull():
                    vobj.Object.Proxy.reload = True
                    vobj.Object.Proxy.execute(vobj.Object)
            else:
                if hasattr(vobj.Object,"ReferenceMode"):
                    if vobj.Object.ReferenceMode == "Transient":
                        vobj.Object.Proxy.reload = False
                        import Part
                        pl = vobj.Object.Placement
                        vobj.Object.Shape = Part.Shape()
                        vobj.Object.Placement = pl


    def onDelete(self,obj,doc):

        if hasattr(self,"timer"):
            self.timer.stop()
            del self.timer
            return True


    def setEdit(self, vobj, mode):
        if mode != 0:
            return None

        taskd = ArchReferenceTaskPanel(vobj.Object)
        FreeCADGui.Control.showDialog(taskd)
        return True


    def unsetEdit(self, vobj, mode):
        if mode != 0:
            return None

        FreeCADGui.Control.closeDialog()
        from DraftGui import todo
        todo.delay(vobj.Proxy.recolorize,vobj)
        return True


    def setupContextMenu(self, vobj, menu):

        actionEdit = QtGui.QAction(translate("Arch", "Edit"),
                                   menu)
        QtCore.QObject.connect(actionEdit,
                               QtCore.SIGNAL("triggered()"),
                               self.edit)
        menu.addAction(actionEdit)

        actionOnReload = QtGui.QAction(QtGui.QIcon(":/icons/view-refresh.svg"),
                                       translate("Arch", "Reload reference"),
                                       menu)
        QtCore.QObject.connect(actionOnReload,
                               QtCore.SIGNAL("triggered()"),
                               self.onReload)
        menu.addAction(actionOnReload)

        actionOnOpen = QtGui.QAction(QtGui.QIcon(":/icons/document-open.svg"),
                                     translate("Arch", "Open reference"),
                                     menu)
        QtCore.QObject.connect(actionOnOpen,
                               QtCore.SIGNAL("triggered()"),
                               self.onOpen)
        menu.addAction(actionOnOpen)


    def edit(self):

        FreeCADGui.ActiveDocument.setEdit(self.Object, 0)


    def onReload(self):

        "reloads the reference object"

        if hasattr(self,"Object") and self.Object:
            self.Object.Proxy.reload = True
            self.Object.touch()
            FreeCAD.ActiveDocument.recompute()


    def onOpen(self):

        "opens the reference file"

        if hasattr(self,"Object") and self.Object:
            if self.Object.File:
                FreeCAD.openDocument(self.Object.File)


    def loadInventor(self, obj):

        "loads an openinventor file and replace the root node of this object"

        filename = obj.Proxy.getFile(obj)
        if not filename:
            return None
        if filename.lower().endswith(".ifc"):
            self.setIFCNode(obj, filename)
            return

        # check inventor contents
        ivstring = self.getInventorString(obj)
        if not ivstring:
            t = translate("Arch","Unable to get lightWeight node for object referenced in")
            FreeCAD.Console.PrintWarning(t+" "+obj.Label+"\n")
            return
        from pivy import coin
        inputnode = coin.SoInput()
        inputnode.setBuffer(ivstring)
        lwnode = coin.SoDB.readAll(inputnode)
        if not isinstance(lwnode,coin.SoSeparator):
            t = translate("Arch","Invalid lightWeight node for object referenced in")
            FreeCAD.Console.PrintError(t+" "+obj.Label+"\n")
            return
        if lwnode.getNumChildren() < 2:
            t = translate("Arch","Invalid lightWeight node for object referenced in")
            FreeCAD.Console.PrintError(t+" "+obj.Label+"\n")
            return
        flatlines = lwnode
        shaded = lwnode.getChild(0)
        wireframe = lwnode.getChild(1)

        # check node contents
        rootnode = obj.ViewObject.RootNode
        if rootnode.getNumChildren() < 3:
            FreeCAD.Console.PrintError(translate("Arch","Invalid root node in")+" "+obj.Label+"\n")
            return
        switch = rootnode.getChild(2)
        if switch.getNumChildren() != 4:
            FreeCAD.Console.PrintError(translate("Arch","Invalid root node in")+" "+obj.Label+"\n")
            return

        # keep a copy of the original nodes
        self.orig_flatlines = switch.getChild(0).copy()
        self.orig_shaded = switch.getChild(1).copy()
        self.orig_wireframe = switch.getChild(2).copy()

        # replace root node of object
        switch.replaceChild(0,flatlines)
        switch.replaceChild(1,shaded)
        switch.replaceChild(2,wireframe)


    def unloadInventor(self,obj):

        "restore original nodes"

        if (not hasattr(self,"orig_flatlines")) or (not self.orig_flatlines):
            return
        if (not hasattr(self,"orig_shaded")) or (not self.orig_shaded):
            return
        if (not hasattr(self,"orig_wireframe")) or (not self.orig_wireframe):
            return

        # check node contents
        rootnode = obj.ViewObject.RootNode
        if rootnode.getNumChildren() < 3:
            FreeCAD.Console.PrintError(translate("Arch","Invalid root node in")+" "+obj.Label+"\n")
            return
        switch = rootnode.getChild(2)
        if switch.getNumChildren() != 4:
            FreeCAD.Console.PrintError(translate("Arch","Invalid root node in")+" "+obj.Label+"\n")
            return

        # replace root node of object
        switch.replaceChild(0,self.orig_flatlines)
        switch.replaceChild(1,self.orig_shaded)
        switch.replaceChild(2,self.orig_wireframe)

        # discard old content
        self.orig_flatlines = None
        self.orig_shaded = None
        self.orig_wireframe = None


    def getInventorString(self,obj):

        "locates and loads an iv file saved together with an object, if existing"

        filename = obj.Proxy.getFile(obj)
        if not filename:
            return None
        part = obj.Part
        if not obj.Part:
            return None
        zdoc = zipfile.ZipFile(filename)
        if not "Document.xml" in zdoc.namelist():
            return None
        ivfile = None
        with zdoc.open("Document.xml") as docf:
            writemode1 = False
            writemode2 = False
            for line in docf:
                line = line.decode("utf8")
                if ("<Object name=" in line) and (part in line):
                    writemode1 = True
                elif writemode1 and ("<Property name=\"SavedInventor\"" in line):
                    writemode1 = False
                    writemode2 = True
                elif writemode2 and ("<FileIncluded file=" in line):
                    n = re.findall(r'file=\"(.*?)\"',line)
                    if n:
                        ivfile = n[0]
                        break
        if not ivfile:
            return None
        if not ivfile in zdoc.namelist():
            return None
        f = zdoc.open(ivfile)
        buf = f.read()
        buf = buf.decode("utf8")
        f.close()
        buf = buf.replace("lineWidth 2","lineWidth "+str(int(obj.ViewObject.LineWidth)))
        return buf


    def setIFCNode(self, obj, filename):

        """Sets the coin node of this object from an IFC file"""

        try:
            from nativeifc import ifc_tools
            from nativeifc import ifc_generator
        except:
            t = translate("Arch","NativeIFC not available - unable to process IFC files")
            FreeCAD.Console.PrintError(t+"\n")
            return
        ifcfile = obj.Proxy.getIfcFile(filename)
        elements = obj.Proxy.getIFCElements(obj, ifcfile)
        node, placement = ifc_generator.generate_coin(ifcfile, elements, cached=True)
        if node:
            ifc_generator.set_representation(obj.ViewObject, node)
            colors = node[0]
            if colors:
                ifc_tools.set_colors(obj, colors)
        else:
            ifc_generator.set_representation(obj.ViewObject, None)
        if placement:
            obj.Placement = placement


class ArchReferenceTaskPanel:


    '''The editmode TaskPanel for Reference objects'''

    def __init__(self,obj):

        self.obj = obj
        self.filename = None
        self.form = QtGui.QWidget()
        self.form.setWindowTitle(translate("Arch","External reference"))
        layout = QtGui.QVBoxLayout(self.form)
        label1 = QtGui.QLabel(translate("Arch","External file")+":")
        layout.addWidget(label1)
        self.fileButton = QtGui.QPushButton(self.form)
        self.openButton = QtGui.QPushButton(self.form)
        self.openButton.setText(translate("Arch","Open"))
        if not self.obj.File:
            self.openButton.setEnabled(False)
        l2 = QtGui.QHBoxLayout()
        layout.addLayout(l2)
        l2.addWidget(self.fileButton)
        l2.addWidget(self.openButton)
        label2 = QtGui.QLabel(translate("Arch","Part to use:"))
        layout.addWidget(label2)
        if self.obj.File:
            self.fileButton.setText(os.path.basename(self.obj.File))
        else:
            self.fileButton.setText(translate("Arch","Choose file..."))
        self.partCombo = QtGui.QComboBox(self.form)
        self.partCombo.setEnabled(False)
        layout.addWidget(self.partCombo)
        if hasattr(self.obj.Proxy,"parts"):
            parts = self.obj.Proxy.parts
        else:
            parts = self.obj.Proxy.getPartsList(self.obj)
        if parts:
            self.partCombo.setEnabled(True)
            sortedkeys = sorted(parts)
            self.partCombo.addItem(translate("Arch","None (Use whole object)"),"")
            for k in sortedkeys:
                self.partCombo.addItem(parts[k][0],k)
            if self.obj.Part:
                if self.obj.Part in sortedkeys:
                    self.partCombo.setCurrentIndex(sortedkeys.index(self.obj.Part))
        else:
            self.partCombo.setEnabled(False)
        QtCore.QObject.connect(self.fileButton, QtCore.SIGNAL("clicked()"), self.chooseFile)
        QtCore.QObject.connect(self.openButton, QtCore.SIGNAL("clicked()"), self.openFile)

    def accept(self):

        from PySide import QtCore

        if self.filename:
            if self.filename != self.obj.File:
                self.obj.File = self.filename
                FreeCAD.ActiveDocument.recompute()
        if self.partCombo.currentText():
            i = self.partCombo.currentIndex()
            if i >= 1:
                if self.partCombo.itemData(i) != self.obj.Part:
                    self.obj.Part = self.partCombo.itemData(i)
            else:
                self.obj.Part = ""
            QtCore.QTimer.singleShot(0,FreeCAD.ActiveDocument.recompute)
        if self.filename and self.obj.Label == "External Reference":
            self.obj.Label = os.path.basename(self.filename)
        FreeCADGui.ActiveDocument.resetEdit()
        return True

    def reject(self):

        FreeCAD.ActiveDocument.recompute()
        FreeCADGui.ActiveDocument.resetEdit()
        return True

    def chooseFile(self):

        loc = QtCore.QDir.homePath()
        if self.obj.File:
            loc = os.path.dirname(self.obj.File)
        filters = "*.FCStd *.dxf"
        # enable IFC support if NativeIFC is present
        try:
            from nativeifc import ifc_tools
        except:
            pass
        else:
            filters += " *.ifc"
        filters = translate("Arch","Reference files")+" ("+filters+")"
        f = QtGui.QFileDialog.getOpenFileName(self.form,
                                              translate("Arch","Choose reference file"),
                                              loc,
                                              filters)
        if f:
            self.filename = f[0]
            self.fileButton.setText(os.path.basename(self.filename))
            parts = self.obj.Proxy.getPartsList(self.obj,self.filename)
            self.partCombo.clear()
            if parts:
                self.partCombo.setEnabled(True)
                sortedkeys = sorted(parts)
                self.partCombo.addItem(translate("Arch","None (Use whole object)"),"")
                for k in sortedkeys:
                    self.partCombo.addItem(parts[k][0],k)
                if self.obj.Part:
                    if self.obj.Part in sortedkeys:
                        self.partCombo.setCurrentIndex(sortedkeys.index(self.obj.Part))
            else:
                self.partCombo.setEnabled(False)

    def openFile(self):

        if self.obj.File:
            if self.obj.File.lower().endswith(".fcstd"):
                FreeCAD.openDocument(self.obj.File)
            else:
                FreeCAD.loadFile(self.obj.File)
            FreeCADGui.Control.closeDialog()
            FreeCADGui.ActiveDocument.resetEdit()






from FreeCAD import Vector, Base, newDocument, closeDocument
import Part
import Sketcher

import unittest

class RegressionTests(unittest.TestCase):

    # pylint: disable=attribute-defined-outside-init

    def setUp(self):
        """Create a document for each test in the test suite"""
        self.Doc = newDocument("PartRegressionTest." + self._testMethodName)
        self.KeepTestDoc = False

    def test_issue_4456(self):
        """
        0004456: Regression : Part.Plane.Intersect does not accept plane as argument
        """
        p1 = Part.Plane()
        p2 = Part.Plane(Vector(0, 0, 0), Vector(1, 0, 0))

        result = p1.intersect(p2)
        line = result.pop()
        self.assertEqual(line.Location, Vector(0, 0, 0))
        self.assertEqual(line.Direction, Vector(0, 1, 0))
        # We should now have empty list...
        with self.assertRaises(IndexError):
            result.pop()

    def test_issue_15735(self):
        """
        15735: Point in sketch as loft profile won't work in dev, but works in stable
        The following test is a simplified version of the issue, but the outcome is the same
        """

        # Arrange
        ArcSketch = self.Doc.addObject("Sketcher::SketchObject", "ArcSketch")
        ArcSketch.Placement = Base.Placement(
            Base.Vector(0.000000, 0.000000, 0.000000),
            Base.Rotation(0.500000, 0.500000, 0.500000, 0.500000),
        )
        ArcSketch.MapMode = "Deactivated"

        geoList = []
        geoList.append(
            Part.ArcOfCircle(
                Part.Circle(
                    Base.Vector(0.000000, 0.000000, 0.000000),
                    Base.Vector(0.000000, 0.000000, 1.000000),
                    10.000000,
                ),
                3.141593,
                6.283185,
            )
        )
        ArcSketch.addGeometry(geoList, False)
        del geoList

        constraintList = []
        ArcSketch.addConstraint(Sketcher.Constraint("Radius", 0, 10.000000))
        constraintList.append(Sketcher.Constraint("Coincident", 0, 3, -1, 1))
        constraintList.append(Sketcher.Constraint("PointOnObject", 0, 2, -1))
        constraintList.append(Sketcher.Constraint("PointOnObject", 0, 1, -1))
        ArcSketch.addConstraint(constraintList)
        del constraintList

        self.Doc.recompute()

        PointSketch = self.Doc.addObject("Sketcher::SketchObject", "PointSketch")
        PointSketch.Placement = Base.Placement(
            Base.Vector(-10.000000, 0.000000, 0.000000),
            Base.Rotation(0.500000, 0.500000, 0.500000, 0.500000),
        )
        PointSketch.MapMode = "Deactivated"

        PointSketch.addGeometry(Part.Point(Base.Vector(0.000000, 0.000000, 0)))

        PointSketch.addConstraint(Sketcher.Constraint("Coincident", 0, 1, -1, 1))

        self.Doc.recompute()

        Loft = self.Doc.addObject("Part::Loft", "Loft")
        Loft.Sections = [
            ArcSketch,
            PointSketch,
        ]
        Loft.Solid = False
        Loft.Ruled = False
        Loft.Closed = False

        # Act
        self.Doc.recompute()

        # Assert
        self.assertTrue(Loft.isValid())
        self.KeepTestDoc = not Loft.isValid()

    def test_OptimalBox(self):
        box = Part.makeBox(1, 1, 1)
        self.assertTrue(box.optimalBoundingBox(True, False).isValid())

    def test_CircularReference(self):

        cube = self.Doc.addObject("Part::Box","Cube")
        cube.setExpression('Length', 'Width + 10mm')
        with self.assertRaises(RuntimeError) as context:
            cube.setExpression('Width', 'Length + 10mm')
        assert "Width reference creates a cyclic dependency." in str(context.exception)

        cube.setExpression('.Placement.Base.x', '.Placement.Base.y + 10mm')
        with self.assertRaises(RuntimeError) as context:
            cube.setExpression('.Placement.Base.y', '.Placement.Base.x + 10mm')
        assert ".Placement.Base.y reference creates a cyclic dependency." in str(context.exception)

        cube.recompute()
        v1 = cube.Placement.Base
        cube.recompute()
        assert cube.Placement.Base.isEqual(v1,1e-6)

    def tearDown(self):
        """Clean up our test, optionally preserving the test document"""
        # This flag allows doing something like this:
        #   self.KeepTestDoc = True
        #   import TestApp
        #   TestApp.Test("TestPartApp.RegressionTests.test_issue_15735")
        # to leave the test document(s) around for further examination in an interactive setting.
        if hasattr(self, "KeepTestDoc") and self.KeepTestDoc:
            return
        closeDocument(self.Doc.Name)



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2024 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

""" Defines a QWidget-derived class for displaying the view selection buttons. """

try:
    import FreeCAD

    translate = FreeCAD.Qt.translate
except ImportError:
    FreeCAD = None

    def translate(_: str, text: str):
        return text


# Get whatever version of PySide we can
try:
    import PySide  # Use the FreeCAD wrapper
except ImportError:
    try:
        import PySide6  # Outside FreeCAD, try Qt6 first

        PySide = PySide6
    except ImportError:
        import PySide2  # Fall back to Qt5 (if this fails, Python will kill this module's import)

        PySide = PySide2

from PySide import QtCore, QtGui, QtWidgets


class WidgetSearch(QtWidgets.QWidget):
    """A widget for selecting the Addon Manager's primary view mode"""

    search_changed = QtCore.Signal(str)

    def __init__(self, parent: QtWidgets.QWidget = None):
        super().__init__(parent)
        self._setup_ui()
        self._setup_connections()
        self.retranslateUi(None)

    def _setup_ui(self):
        self.horizontal_layout = QtWidgets.QHBoxLayout()
        self.horizontal_layout.setContentsMargins(0, 0, 0, 0)
        self.filter_line_edit = QtWidgets.QLineEdit(self)
        self.filter_line_edit.setClearButtonEnabled(True)
        self.horizontal_layout.addWidget(self.filter_line_edit)
        self.filter_validity_label = QtWidgets.QLabel(self)
        self.horizontal_layout.addWidget(self.filter_validity_label)
        self.filter_validity_label.hide()  # This widget starts hidden
        self.setLayout(self.horizontal_layout)

    def _setup_connections(self):
        self.filter_line_edit.textChanged.connect(self.set_text_filter)

    def set_text_filter(self, text_filter: str) -> None:
        """Set the current filter. If the filter is valid, this will emit a filter_changed
        signal. text_filter may be regular expression."""

        if text_filter:
            test_regex = QtCore.QRegularExpression(text_filter)
            if test_regex.isValid():
                self.filter_validity_label.setToolTip(
                    translate("AddonsInstaller", "Filter is valid")
                )
                icon = QtGui.QIcon.fromTheme("ok", QtGui.QIcon(":/icons/edit_OK.svg"))
                self.filter_validity_label.setPixmap(icon.pixmap(16, 16))
            else:
                self.filter_validity_label.setToolTip(
                    translate("AddonsInstaller", "Filter regular expression is invalid")
                )
                icon = QtGui.QIcon.fromTheme("cancel", QtGui.QIcon(":/icons/edit_Cancel.svg"))
                self.filter_validity_label.setPixmap(icon.pixmap(16, 16))
            self.filter_validity_label.show()
        else:
            self.filter_validity_label.hide()
        self.search_changed.emit(text_filter)

    def retranslateUi(self, _):
        self.filter_line_edit.setPlaceholderText(
            QtCore.QCoreApplication.translate("AddonsInstaller", "Search...", None)
        )



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022-2024 FreeCAD Project Association                   *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

""" Defines a QWidget-derived class for displaying the cache load status. """

try:
    import FreeCAD

    translate = FreeCAD.Qt.translate
except ImportError:
    FreeCAD = None

    def translate(_: str, text: str):
        return text


# Get whatever version of PySide we can
try:
    import PySide  # Use the FreeCAD wrapper
except ImportError:
    try:
        import PySide6  # Outside FreeCAD, try Qt6 first

        PySide = PySide6
    except ImportError:
        import PySide2  # Fall back to Qt5 (if this fails, Python will kill this module's import)

        PySide = PySide2

from PySide import QtCore, QtGui, QtWidgets

_TOTAL_INCREMENTS = 1000


class WidgetProgressBar(QtWidgets.QWidget):
    """A multipart progress bar widget, including a stop button and a status label. Defaults to a
    single range with 100 increments, but can be configured with any number of major and minor
    ranges. Clicking the stop button will emit a signal, but does not otherwise affect the
    widget."""

    stop_clicked = QtCore.Signal()

    def __init__(self, parent: QtWidgets.QWidget = None):
        super().__init__(parent)
        self.vertical_layout = None
        self.horizontal_layout = None
        self.progress_bar = None
        self.status_label = None
        self.stop_button = None
        self._setup_ui()

    def _setup_ui(self):
        self.vertical_layout = QtWidgets.QVBoxLayout(self)
        self.horizontal_layout = QtWidgets.QHBoxLayout()
        self.progress_bar = QtWidgets.QProgressBar(self)
        self.status_label = QtWidgets.QLabel(self)
        self.stop_button = QtWidgets.QToolButton(self)
        self.progress_bar.setMaximum(_TOTAL_INCREMENTS)
        self.stop_button.clicked.connect(self.stop_clicked)
        self.stop_button.setIcon(
            QtGui.QIcon.fromTheme("stop", QtGui.QIcon(":/icons/debug-stop.svg"))
        )
        self.vertical_layout.addLayout(self.horizontal_layout)
        self.vertical_layout.addWidget(self.status_label)
        self.horizontal_layout.addWidget(self.progress_bar)
        self.horizontal_layout.addWidget(self.stop_button)
        self.vertical_layout.setContentsMargins(0, 0, 0, 0)
        self.horizontal_layout.setContentsMargins(0, 0, 0, 0)

    def set_status(self, status: str):
        self.status_label.setText(status)

    def set_value(self, value: int):
        self.progress_bar.setValue(value)



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2024 The FreeCAD Project Association AISBL              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

""" A Qt Widget for displaying Addon README information """

import FreeCAD
from Addon import Addon
import addonmanager_utilities as utils

from enum import IntEnum, Enum, auto
from html.parser import HTMLParser
from typing import Optional

import NetworkManager

translate = FreeCAD.Qt.translate

from PySide import QtCore, QtGui


class ReadmeDataType(IntEnum):
    PlainText = 0
    Markdown = 1
    Html = 2


class ReadmeController(QtCore.QObject):
    """A class that can provide README data from an Addon, possibly loading external resources such
    as images"""

    def __init__(self, widget):
        super().__init__()
        NetworkManager.InitializeNetworkManager()
        NetworkManager.AM_NETWORK_MANAGER.completed.connect(self._download_completed)
        self.readme_request_index = 0
        self.resource_requests = {}
        self.resource_failures = []
        self.url = ""
        self.readme_data = None
        self.readme_data_type = None
        self.addon: Optional[Addon] = None
        self.stop = True
        self.widget = widget
        self.widget.load_resource.connect(self.loadResource)
        self.widget.follow_link.connect(self.follow_link)

    def set_addon(self, repo: Addon):
        """Set which Addon's information is displayed"""

        self.addon = repo
        self.stop = False
        self.readme_data = None
        if self.addon.repo_type == Addon.Kind.MACRO:
            self.url = self.addon.macro.wiki
            if not self.url:
                self.url = self.addon.macro.url
            if not self.url:
                self.widget.setText(
                    translate(
                        "AddonsInstaller",
                        "Loading info for {} from the FreeCAD Macro Recipes wiki...",
                    ).format(self.addon.display_name, self.url)
                )
                return
        else:
            self.url = utils.get_readme_url(repo)
        self.widget.setUrl(self.url)

        self.widget.setText(
            translate("AddonsInstaller", "Loading page for {} from {}...").format(
                self.addon.display_name, self.url
            )
        )
        self.readme_request_index = NetworkManager.AM_NETWORK_MANAGER.submit_unmonitored_get(
            self.url
        )

    def _download_completed(self, index: int, code: int, data: QtCore.QByteArray) -> None:
        """Callback for handling a completed README file download."""
        if index == self.readme_request_index:
            if code == 200:  # HTTP success
                self._process_package_download(data.data().decode("utf-8"))
            else:
                self.widget.setText(
                    translate(
                        "AddonsInstaller",
                        "Failed to download data from {} -- received response code {}.",
                    ).format(self.url, code)
                )
        elif index in self.resource_requests:
            if code == 200:
                self._process_resource_download(self.resource_requests[index], data.data())
            else:
                FreeCAD.Console.PrintLog(f"Failed to load {self.resource_requests[index]}\n")
                self.resource_failures.append(self.resource_requests[index])
            del self.resource_requests[index]
            if not self.resource_requests:
                if self.readme_data:
                    if self.readme_data_type == ReadmeDataType.Html:
                        self.widget.setHtml(self.readme_data)
                    elif self.readme_data_type == ReadmeDataType.Markdown:
                        self.widget.setMarkdown(self.readme_data)
                    else:
                        self.widget.setText(self.readme_data)
                else:
                    self.set_addon(self.addon)  # Trigger a reload of the page now with resources

    def _process_package_download(self, data: str):
        if self.addon.repo_type == Addon.Kind.MACRO:
            parser = WikiCleaner()
            parser.feed(data)
            self.readme_data = parser.final_html
            self.readme_data_type = ReadmeDataType.Html
            self.widget.setHtml(parser.final_html)
        else:
            self.readme_data = data
            self.readme_data_type = ReadmeDataType.Markdown
            self.widget.setMarkdown(data)

    def _process_resource_download(self, resource_name: str, resource_data: bytes):
        image = QtGui.QImage.fromData(resource_data)
        self.widget.set_resource(resource_name, image)

    def loadResource(self, full_url: str):
        if full_url not in self.resource_failures:
            index = NetworkManager.AM_NETWORK_MANAGER.submit_unmonitored_get(full_url)
            self.resource_requests[index] = full_url

    def cancel_resource_loading(self):
        self.stop = True
        for request in self.resource_requests:
            NetworkManager.AM_NETWORK_MANAGER.abort(request)
        self.resource_requests.clear()

    def follow_link(self, url: str) -> None:
        final_url = url
        if not url.startswith("http"):
            if url.endswith(".md"):
                final_url = self._create_markdown_url(url)
            else:
                final_url = self._create_full_url(url)
        FreeCAD.Console.PrintLog(f"Loading {final_url} in the system browser")
        QtGui.QDesktopServices.openUrl(final_url)

    def _create_full_url(self, url: str) -> str:
        if url.startswith("http"):
            return url
        if not self.url:
            return url
        lhs, slash, _ = self.url.rpartition("/")
        return lhs + slash + url

    def _create_markdown_url(self, file: str) -> str:
        base_url = utils.get_readme_html_url(self.addon)
        lhs, slash, _ = base_url.rpartition("/")
        return lhs + slash + file


class WikiCleaner(HTMLParser):
    """This HTML parser cleans up FreeCAD Macro Wiki Page for display in a
    QTextBrowser widget (which does not deal will with tables used as formatting,
    etc.) It strips out any tables, and extracts the mw-parser-output div as the only
    thing that actually gets displayed. It also discards anything inside the [edit]
    spans that litter wiki output."""

    class State(Enum):
        BeforeMacroContent = auto()
        InMacroContent = auto()
        InTable = auto()
        InEditSpan = auto()
        AfterMacroContent = auto()

    def __init__(self):
        super().__init__()
        self.depth_in_div = 0
        self.depth_in_span = 0
        self.depth_in_table = 0
        self.final_html = "<html><body>"
        self.previous_state = WikiCleaner.State.BeforeMacroContent
        self.state = WikiCleaner.State.BeforeMacroContent

    def handle_starttag(self, tag: str, attrs):
        if tag == "div":
            self.handle_div_start(attrs)
        elif tag == "span":
            self.handle_span_start(attrs)
        elif tag == "table":
            self.handle_table_start(attrs)
        else:
            if self.state == WikiCleaner.State.InMacroContent:
                self.add_tag_to_html(tag, attrs)

    def handle_div_start(self, attrs):
        for name, value in attrs:
            if name == "class" and value == "mw-parser-output":
                self.previous_state = self.state
                self.state = WikiCleaner.State.InMacroContent
        if self.state == WikiCleaner.State.InMacroContent:
            self.depth_in_div += 1
            self.add_tag_to_html("div", attrs)

    def handle_span_start(self, attrs):
        for name, value in attrs:
            if name == "class" and value == "mw-editsection":
                self.previous_state = self.state
                self.state = WikiCleaner.State.InEditSpan
                break
        if self.state == WikiCleaner.State.InEditSpan:
            self.depth_in_span += 1
        elif WikiCleaner.State.InMacroContent:
            self.add_tag_to_html("span", attrs)

    def handle_table_start(self, unused):
        if self.state != WikiCleaner.State.InTable:
            self.previous_state = self.state
            self.state = WikiCleaner.State.InTable
        self.depth_in_table += 1

    def add_tag_to_html(self, tag, attrs=None):
        self.final_html += f"<{tag}"
        if attrs:
            self.final_html += " "
            for attr, value in attrs:
                self.final_html += f"{attr}='{value}'"
        self.final_html += ">\n"

    def handle_endtag(self, tag):
        if tag == "table":
            self.handle_table_end()
        elif tag == "span":
            self.handle_span_end()
        elif tag == "div":
            self.handle_div_end()
        else:
            if self.state == WikiCleaner.State.InMacroContent:
                self.add_tag_to_html(f"/{tag}")

    def handle_span_end(self):
        if self.state == WikiCleaner.State.InEditSpan:
            self.depth_in_span -= 1
            if self.depth_in_span <= 0:
                self.depth_in_span = 0
                self.state = self.previous_state
        else:
            self.add_tag_to_html(f"/span")

    def handle_div_end(self):
        if self.state == WikiCleaner.State.InMacroContent:
            self.depth_in_div -= 1
            if self.depth_in_div <= 0:
                self.depth_in_div = 0
                self.state = WikiCleaner.State.AfterMacroContent
                self.final_html += "</body></html>"
        else:
            self.add_tag_to_html(f"/div")

    def handle_table_end(self):
        if self.state == WikiCleaner.State.InTable:
            self.depth_in_table -= 1
            if self.depth_in_table <= 0:
                self.depth_in_table = 0
                self.state = self.previous_state

    def handle_data(self, data):
        if self.state == WikiCleaner.State.InMacroContent:
            self.final_html += data



# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2023 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
#*   You should have received a copy of the GNU Library General Public     *
#*   License along with this program; if not, write to the Free Software   *
#*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
#*   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""This NativeIFC module deals with properties and property sets"""


import re
import FreeCAD
from nativeifc import ifc_tools


def has_psets(obj):
    """Returns True if an object has attached psets"""

    element = ifc_tools.get_ifc_element(obj)
    psets = getattr(element, "IsDefinedBy", [])
    if psets and [p for p in psets if p.is_a("IfcRelDefinesByProperties")]:
        # TODO verify too if these psets are not already there
        return True
    psets = getattr(element, "HasProperties", [])
    if psets:
        return True
    return False


def get_psets(element):
    """Returns a dictionary of dictionaries representing the
    properties of an element in the form:
    { pset_name : { property_name : IfcType(value), ... }, ... }"""

    result = {}
    psets = getattr(element, "IsDefinedBy", [])
    psets = [p for p in psets if p.is_a("IfcRelDefinesByProperties")]
    psets = [p.RelatingPropertyDefinition for p in psets]
    if not psets:
        psets = getattr(element, "HasProperties", [])
    if not psets:
        return result
    for pset in psets:
        pset_dict = {}
        if pset.is_a("IfcPropertySet") and pset.HasProperties:
            for prop in pset.HasProperties:
                pset_dict[prop.Name] = str(prop.NominalValue)
        if pset.is_a("IfcMaterialProperties") and pset.Properties:
            for prop in pset.Properties:
                pset_dict[prop.Name] = str(prop.NominalValue)
        elif pset.is_a("IfcElementQuantity"):
            # TODO implement quantities
            pass
        if pset_dict:
            result[pset.Name] = pset_dict
    return result


def get_pset(psetname, element):
    """Returns an IfcPropertySet with the given name"""

    psets = getattr(element, "IsDefinedBy", [])
    psets = [p for p in psets if p.is_a("IfcRelDefinesByProperties")]
    for p in psets:
        pset = p.RelatingPropertyDefinition
        if pset.Name == psetname:
            return pset
    return None


def show_psets(obj):
    """Shows the psets attached to the given object as properties"""

    element = ifc_tools.get_ifc_element(obj)
    if not element:
        return
    psets = get_psets(element)
    for gname, pset in psets.items():
        for pname, pvalue in pset.items():
            oname = pname
            ptype, value = pvalue.split("(", 1)
            value = value.strip(")")
            value = value.strip("'")
            pname = re.sub(r"[^0-9a-zA-Z]+", "", pname)
            if pname[0].isdigit():
                pname = "_" + pname
            ttip = (
                ptype + ":" + oname
            )  # setting IfcType:PropName as a tooltip to desambiguate
            while pname in obj.PropertiesList:
                # print("DEBUG: property", pname, "(", value, ") already exists in", obj.Label)
                pname += "_"
            if ptype in [
                "IfcPositiveLengthMeasure",
                "IfcLengthMeasure",
                "IfcNonNegativeLengthMeasure",
            ]:
                obj.addProperty("App::PropertyDistance", pname, gname, ttip)
            elif ptype in ["IfcVolumeMeasure"]:
                obj.addProperty("App::PropertyVolume", pname, gname, ttip)
            elif ptype in ["IfcPositivePlaneAngleMeasure", "IfcPlaneAngleMeasure"]:
                obj.addProperty("App::PropertyAngle", pname, gname, ttip)
                value = float(value)
                while value > 360:
                    value = value - 360
            elif ptype in ["IfcMassMeasure"]:
                obj.addProperty("App::PropertyMass", pname, gname, ttip)
            elif ptype in ["IfcAreaMeasure"]:
                obj.addProperty("App::PropertyArea", pname, gname, ttip)
            elif ptype in ["IfcCountMeasure", "IfcInteger"]:
                obj.addProperty("App::PropertyInteger", pname, gname, ttip)
                value = int(value.strip("."))
            elif ptype in ["IfcReal"]:
                obj.addProperty("App::PropertyFloat", pname, gname, ttip)
                value = float(value)
            elif ptype in ["IfcBoolean", "IfcLogical"]:
                obj.addProperty("App::PropertyBool", pname, gname, ttip)
                if value in [".T."]:
                    value = True
                else:
                    value = False
            elif ptype in [
                "IfcDateTime",
                "IfcDate",
                "IfcTime",
                "IfcDuration",
                "IfcTimeStamp",
            ]:
                obj.addProperty("App::PropertyTime", pname, gname, ttip)
            else:
                obj.addProperty("App::PropertyString", pname, gname, ttip)
            # print("DEBUG: setting",pname, ptype, value)
            setattr(obj, pname, value)


def edit_pset(obj, prop, value=None):
    """Edits the corresponding property"""

    pset = obj.getGroupOfProperty(prop)
    ttip = obj.getDocumentationOfProperty(prop)
    if value is None:
        value = getattr(obj, prop)
    ifcfile = ifc_tools.get_ifcfile(obj)
    element = ifc_tools.get_ifc_element(obj)
    pset_exist = get_psets(element)
    if ttip.startswith("Ifc") and ":" in ttip:
        target_prop = ttip.split(":", 1)[-1]
    else:
        # no tooltip set - try to build a name
        prop = prop.rstrip("_")
        prop_uncamel = re.sub(r"(\w)([A-Z])", r"\1 \2", prop)
        prop_unslash = re.sub(r"(\w)([A-Z])", r"\1\/\2", prop)
        target_prop = None
    if pset in pset_exist:
        if not target_prop:
            if prop in pset_exist[pset]:
                target_prop = prop
            elif prop_uncamel in pset_exist[pset]:
                target_prop = prop_uncamel
            elif prop_unslash in pset_exist[pset]:
                target_prop = prop_unslash
        if target_prop:
            value_exist = pset_exist[pset][target_prop].split("(", 1)[1][:-1].strip("'")
            if value_exist in [".F.", ".U."]:
                value_exist = False
            elif value_exist in [".T."]:
                value_exist = True
            elif isinstance(value, int):
                value_exist = int(value_exist.strip("."))
            elif isinstance(value, float):
                value_exist = float(value_exist)
            elif isinstance(value, FreeCAD.Units.Quantity):
                if value.Unit.Type == "Angle":
                    value_exist = float(value_exist)
                    while value_exist > 360:
                        value_exist = value_exist - 360
                value_exist = FreeCAD.Units.Quantity(float(value_exist), value.Unit)
            if value == value_exist:
                return False
            else:
                FreeCAD.Console.PrintLog(
                    "IFC: property changed for "
                    + obj.Label
                    + " ("
                    + str(obj.StepId)
                    + ") : "
                    + str(target_prop)
                    + " : "
                    + str(value)
                    + " ("
                    + str(type(value))
                    + ") -> "
                    + str(value_exist)
                    + " ("
                    + str(type(value_exist))
                    + ")\n"
                )
        pset = get_pset(pset, element)
    else:
        pset = ifc_tools.api_run("pset.add_pset", ifcfile, product=element, name=pset)
    if not target_prop:
        target_prop = prop
    ifc_tools.api_run(
        "pset.edit_pset", ifcfile, pset=pset, properties={target_prop: value}
    )
    # TODO manage quantities
    return True


def load_psets(obj):
    """Recursively loads psets of child objects"""

    show_psets(obj)
    if isinstance(obj, FreeCAD.DocumentObject):
        group = obj.Group
    else:
        group = obj.Objects
    for child in group:
        load_psets(child)


def add_pset(obj, psetname):
    """Adds a pset with the given name to the given object"""

    ifcfile = ifc_tools.get_ifcfile(obj)
    element = ifc_tools.get_ifc_element(obj)
    if ifcfile and element:
        pset = ifc_tools.api_run(
            "pset.add_pset", ifcfile, product=element, name=psetname
        )
        return pset


def add_property(ifcfile, pset, name, value=""):
    """Adds a property with the given name to the given pset. The type is deduced from
    the value: string is IfcLabel, True/False is IfcBoolean, number is IfcLengthMeasure.
    To force a certain type, value can also be an IFC element such as IfcLabel"""

    ifc_tools.api_run("pset.edit_pset", ifcfile, pset=pset, properties={name: value})



# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2024 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""BIM Panel-related Arch_"""


import os
import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate
PARAMS = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/BIM")


class Arch_Pipe:

    "the Arch Pipe command definition"

    def GetResources(self):

        return {'Pixmap'  : 'Arch_Pipe',
                'MenuText': QT_TRANSLATE_NOOP("Arch_Pipe","Pipe"),
                'Accel': "P, I",
                'ToolTip': QT_TRANSLATE_NOOP("Arch_Pipe","Creates a pipe object from a given Wire or Line")}

    def IsActive(self):

        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):

        s = FreeCADGui.Selection.getSelection()
        if s:
            for obj in s:
                if hasattr(obj,'Shape'):
                    if len(obj.Shape.Wires) == 1:
                        FreeCAD.ActiveDocument.openTransaction(translate("Arch","Create Pipe"))
                        FreeCADGui.addModule("Arch")
                        FreeCADGui.addModule("Draft")
                        FreeCADGui.doCommand("obj = Arch.makePipe(FreeCAD.ActiveDocument."+obj.Name+")")
                        FreeCADGui.doCommand("Draft.autogroup(obj)")
                        FreeCAD.ActiveDocument.commitTransaction()
        else:
            FreeCAD.ActiveDocument.openTransaction(translate("Arch","Create Pipe"))
            FreeCADGui.addModule("Arch")
            FreeCADGui.addModule("Draft")
            FreeCADGui.doCommand("obj = Arch.makePipe()")
            FreeCADGui.doCommand("Draft.autogroup(obj)")
            FreeCAD.ActiveDocument.commitTransaction()
        FreeCAD.ActiveDocument.recompute()


class Arch_PipeConnector:

    "the Arch Pipe command definition"

    def GetResources(self):

        return {'Pixmap'  : 'Arch_PipeConnector',
                'MenuText': QT_TRANSLATE_NOOP("Arch_PipeConnector","Connector"),
                'Accel': "P, C",
                'ToolTip': QT_TRANSLATE_NOOP("Arch_PipeConnector","Creates a connector between 2 or 3 selected pipes")}

    def IsActive(self):

        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):

        import Draft
        s = FreeCADGui.Selection.getSelection()
        if not (len(s) in [2,3]):
            FreeCAD.Console.PrintError(translate("Arch","Please select exactly 2 or 3 Pipe objects")+"\n")
            return
        o = "["
        for obj in s:
            if Draft.getType(obj) != "Pipe":
                FreeCAD.Console.PrintError(translate("Arch","Please select only Pipe objects")+"\n")
                return
            o += "FreeCAD.ActiveDocument."+obj.Name+","
        o += "]"
        FreeCAD.ActiveDocument.openTransaction(translate("Arch","Create Connector"))
        FreeCADGui.addModule("Arch")
        FreeCADGui.addModule("Draft")
        FreeCADGui.doCommand("obj = Arch.makePipeConnector("+o+")")
        FreeCADGui.doCommand("Draft.autogroup(obj)")
        FreeCAD.ActiveDocument.commitTransaction()
        FreeCAD.ActiveDocument.recompute()


class Arch_PipeGroupCommand:

    def GetCommands(self):
        return tuple(['Arch_Pipe','Arch_PipeConnector'])
    def GetResources(self):
        return { 'MenuText': QT_TRANSLATE_NOOP("Arch_PipeTools",'Pipe tools'),
                 'ToolTip': QT_TRANSLATE_NOOP("Arch_PipeTools",'Pipe tools')
               }
    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v


FreeCADGui.addCommand('Arch_Pipe',Arch_Pipe())
FreeCADGui.addCommand('Arch_PipeConnector',Arch_PipeConnector())
FreeCADGui.addCommand('Arch_PipeTools', Arch_PipeGroupCommand())



# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2024 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""BIM Panel-related Arch_"""


import os
import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate
PARAMS = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/BIM")


class Arch_Profile:

    """The FreeCAD Arch_Profile command definition"""

    def GetResources(self):

        return {'Pixmap'  : 'Arch_Profile',
                'MenuText': QT_TRANSLATE_NOOP("Arch_Profile","Profile"),
                'Accel': "P, F",
                'ToolTip': QT_TRANSLATE_NOOP("Arch_Profile","Creates a profile")}

    def IsActive(self):

        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):

        import ArchProfile
        self.Profile = None
        self.Categories = []
        self.Presets = ArchProfile.readPresets()
        for pre in self.Presets:
            if pre[1] not in self.Categories:
                self.Categories.append(pre[1])
        FreeCADGui.Snapper.getPoint(callback=self.getPoint,extradlg=[self.taskbox()],title=translate("Arch","Create profile"))

    def taskbox(self):

        "sets up a taskbox widget"

        from PySide import QtGui
        from draftutils import params
        w = QtGui.QWidget()
        ui = FreeCADGui.UiLoader()
        w.setWindowTitle(translate("Arch","Profile settings"))
        grid = QtGui.QGridLayout(w)

        # categories box
        labelc = QtGui.QLabel(translate("Arch","Category"))
        self.vCategory = QtGui.QComboBox()
        self.vCategory.addItems([" "] + self.Categories)
        grid.addWidget(labelc,1,0,1,1)
        grid.addWidget(self.vCategory,1,1,1,1)

        # presets box
        labelp = QtGui.QLabel(translate("Arch","Preset"))
        self.vPresets = QtGui.QComboBox()
        self.pSelect = [None]
        fpresets = [" "]
        self.vPresets.addItems(fpresets)
        grid.addWidget(labelp,2,0,1,1)
        grid.addWidget(self.vPresets,2,1,1,1)

        # connect slots
        self.vCategory.currentIndexChanged.connect(self.setCategory)
        self.vPresets.currentIndexChanged.connect(self.setPreset)

        # restore preset
        stored = params.get_param_arch("StructurePreset")
        if stored:
            if ";" in stored:
                stored = stored.split(";")
                if len(stored) >= 3:
                    if stored[1] in self.Categories:
                        self.vCategory.setCurrentIndex(1+self.Categories.index(stored[1]))
                        self.setCategory(1+self.Categories.index(stored[1]))
                        ps = [p[2] for p in self.pSelect]
                        if stored[2] in ps:
                            self.vPresets.setCurrentIndex(ps.index(stored[2]))
        return w

    def getPoint(self,point=None,obj=None):

        "this function is called by the snapper when it has a 3D point"

        if not point:
            return
        if not self.Profile:
            return
        pt = "FreeCAD.Vector("+str(point.x)+","+str(point.y)+","+str(point.z)+")"
        FreeCAD.ActiveDocument.openTransaction(translate("Arch","Create Profile"))
        FreeCADGui.addModule("Arch")
        FreeCADGui.doCommand('p = Arch.makeProfile('+str(self.Profile)+')')
        FreeCADGui.addModule('WorkingPlane')
        FreeCADGui.doCommand('p.Placement = WorkingPlane.get_working_plane().get_placement()')
        FreeCADGui.doCommand('p.Placement.Base = ' + pt)
        FreeCADGui.addModule("Draft")
        FreeCADGui.doCommand("Draft.autogroup(p)")
        FreeCAD.ActiveDocument.commitTransaction()
        FreeCAD.ActiveDocument.recompute()

    def setCategory(self,i):

        from draftutils import params
        self.vPresets.clear()
        if i == 0:
            self.pSelect = [None]
            self.vPresets.addItems([" "])
            params.set_param_arch("StructurePreset","")
        else:
            self.pSelect = [p for p in self.Presets if p[1] == self.Categories[i-1]]
            fpresets = []
            for p in self.pSelect:
                f = FreeCAD.Units.Quantity(p[4],FreeCAD.Units.Length).getUserPreferred()
                d = params.get_param("Decimals",path="Units")
                s1 = str(round(p[4]/f[1],d))
                s2 = str(round(p[5]/f[1],d))
                s3 = str(f[2])
                fpresets.append(p[2]+" ("+s1+"x"+s2+s3+")")
            self.vPresets.addItems(fpresets)
            self.setPreset(0)

    def setPreset(self,i):

        from draftutils import params
        self.Profile = None
        elt = self.pSelect[i]
        if elt:
            p=elt[0]-1 # Presets indexes are 1-based
            self.Profile = self.Presets[p]
            params.set_param_arch("StructurePreset",";".join([str(i) for i in self.Profile]))


FreeCADGui.addCommand('Arch_Profile',Arch_Profile())



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022-2023 FreeCAD Project Association                   *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

from datetime import date, timedelta
import hashlib
import os

import addonmanager_freecad_interface as fci
import addonmanager_utilities as utils

translate = fci.translate


def local_cache_needs_update() -> bool:
    """Determine whether we need to update the cache, based on user preference, and previous
    cache update status. Returns either True or False."""

    if not _cache_exists():
        return True

    if _last_update_was_interrupted(reset_status=True):
        return True

    if _custom_repo_list_changed():
        return True

    # Figure out our cache update frequency: there is a combo box in the preferences dialog
    # with three options: never, daily, and weekly.
    days_between_updates = _days_between_updates()
    pref = fci.ParamGet("User parameter:BaseApp/Preferences/Addons")
    last_cache_update_string = pref.GetString("LastCacheUpdate", "never")

    if last_cache_update_string == "never":
        return True
    elif days_between_updates > 0:
        last_cache_update = date.fromisoformat(last_cache_update_string)
        delta_update = timedelta(days=days_between_updates)
        if date.today() >= last_cache_update + delta_update:
            return True
    elif days_between_updates == 0:
        return True

    return False


def _days_between_updates() -> int:
    pref = fci.ParamGet("User parameter:BaseApp/Preferences/Addons")
    update_frequency = pref.GetInt("UpdateFrequencyComboEntry", 0)
    if update_frequency == 0:
        return -1
    elif update_frequency == 1:
        return 1
    elif update_frequency == 2:
        return 7
    else:
        return 0


def _cache_exists() -> bool:
    cache_path = fci.getUserCachePath()
    am_path = os.path.join(cache_path, "AddonManager")
    return os.path.exists(am_path)


def _last_update_was_interrupted(reset_status: bool) -> bool:
    flag_file = utils.get_cache_file_name("CACHE_UPDATE_INTERRUPTED")
    if os.path.exists(flag_file):
        if reset_status:
            os.remove(flag_file)
        fci.Console.PrintMessage(
            translate(
                "AddonsInstaller",
                "Previous cache process was interrupted, restarting...\n",
            )
        )
        return True


def _custom_repo_list_changed() -> bool:
    pref = fci.ParamGet("User parameter:BaseApp/Preferences/Addons")
    stored_hash = pref.GetString("CustomRepoHash", "")
    custom_repos = pref.GetString("CustomRepositories", "")
    if custom_repos:
        hasher = hashlib.sha1()
        hasher.update(custom_repos.encode("utf-8"))
        new_hash = hasher.hexdigest()
    else:
        new_hash = ""
    if new_hash != stored_hash:
        pref.SetString("CustomRepoHash", new_hash)
        fci.Console.PrintMessage(
            translate(
                "AddonsInstaller",
                "Custom repo list changed, forcing recache...\n",
            )
        )
        return True
    return False



import unittest

import FreeCAD as App


class ColorTransparencyTest(unittest.TestCase):

    def setUp(self):
        self._doc = App.newDocument()
        self._pg = App.ParamGet('User parameter:BaseApp/Preferences/View')
        self._backup_default_transparency = self._pg.GetInt('DefaultShapeTransparency')
        self._backup_default_shapecolor = self._pg.GetUnsigned('DefaultShapeColor')


    def tearDown(self):
        App.closeDocument(self._doc.Name)
        self._pg.SetInt('DefaultShapeTransparency', self._backup_default_transparency)
        self._pg.SetUnsigned('DefaultShapeColor', self._backup_default_shapecolor)


    def test_default_shape_transparency(self):
        """
        related: https://github.com/FreeCAD/FreeCAD/pull/11866
        related: https://github.com/FreeCAD/FreeCAD/pull/11586
        """
        transparency = 70
        self._pg.SetInt('DefaultShapeTransparency', transparency)
        obj = self._doc.addObject('Part::Box')
        assert obj.ViewObject.Transparency == transparency
        obj.ViewObject.ShapeAppearance[0].DiffuseColor = (0.5, 0.0, 0.0)

        self.assertEqual(obj.ViewObject.Transparency, transparency,
            'transparency was unexpectedly changed to {} when changing the color.'.format(
            obj.ViewObject.Transparency))


    def test_default_shape_color(self):
        """
        related: https://github.com/FreeCAD/FreeCAD/pull/11866
        """

        """
        This test isn't currently valid as it draws from the hard coded default material.

        The preference editor doesn't allow for setting transparencies. The default value
        of 0 corresponds to a fully transparent color, which is not desirable. It changes
        the transparency when loading to 1.0
        """
        
        self._pg.SetUnsigned('DefaultShapeColor', 0xff000000)  # red
        obj = self._doc.addObject('Part::Box')

        self.assertEqual(obj.ViewObject.ShapeAppearance[0].DiffuseColor, (1.0, 0.0, 0.0, 1.0),
            'default shape color was not set correctly')
        self.assertEqual(obj.ViewObject.ShapeMaterial.DiffuseColor, (1.0, 0.0, 0.0, 1.0),
            'default material color was not set correctly')


    def test_app_plane_transparency(self):
        """
        related: https://github.com/FreeCAD/FreeCAD/pull/12064
        """
        self._pg.SetInt('DefaultShapeTransparency', 70)
        obj = self._doc.addObject('App::Origin')
        t = self._doc.findObjects('App::Plane')[0].ViewObject.Transparency

        self.assertEqual(t, 0,
            'transparency of App::Plane object is {} instead of 0'.format(t))




# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022-2024 The FreeCAD Project Association AISBL         *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

""" Provides the PackageDetails widget. """

import os
from typing import Optional

from PySide import QtCore, QtWidgets

import addonmanager_freecad_interface as fci

import addonmanager_utilities as utils
from addonmanager_metadata import (
    Version,
    get_first_supported_freecad_version,
    get_branch_from_metadata,
)
from addonmanager_workers_startup import GetMacroDetailsWorker, CheckSingleUpdateWorker
from addonmanager_git import GitManager, NoGitFound
from Addon import Addon
from change_branch import ChangeBranchDialog
from addonmanager_readme_controller import ReadmeController
from Widgets.addonmanager_widget_package_details_view import UpdateInformation, WarningFlags

translate = fci.translate


class PackageDetailsController(QtCore.QObject):
    """Manages the display of the package README information."""

    back = QtCore.Signal()
    install = QtCore.Signal(Addon)
    uninstall = QtCore.Signal(Addon)
    update = QtCore.Signal(Addon)
    execute = QtCore.Signal(Addon)
    update_status = QtCore.Signal(Addon)

    def __init__(self, widget=None):
        super().__init__()
        self.ui = widget
        self.readme_controller = ReadmeController(self.ui.readme_browser)
        self.worker = None
        self.addon = None
        self.update_check_thread = None
        self.original_disabled_state = None
        self.original_status = None
        self.check_for_update_worker = None
        try:
            self.git_manager = GitManager()
        except NoGitFound:
            self.git_manager = None

        self.ui.button_bar.back.clicked.connect(self.back.emit)
        self.ui.button_bar.run_macro.clicked.connect(lambda: self.execute.emit(self.addon))
        self.ui.button_bar.install.clicked.connect(lambda: self.install.emit(self.addon))
        self.ui.button_bar.uninstall.clicked.connect(lambda: self.uninstall.emit(self.addon))
        self.ui.button_bar.update.clicked.connect(lambda: self.update.emit(self.addon))
        self.ui.button_bar.change_branch.clicked.connect(self.change_branch_clicked)
        self.ui.button_bar.enable.clicked.connect(self.enable_clicked)
        self.ui.button_bar.disable.clicked.connect(self.disable_clicked)

    def show_repo(self, repo: Addon) -> None:
        """The main entry point for this class, shows the package details and related buttons
        for the provided repo."""
        self.addon = repo
        self.readme_controller.set_addon(repo)
        self.original_disabled_state = self.addon.is_disabled()
        if repo is not None:
            self.ui.button_bar.show()
        else:
            self.ui.button_bar.hide()

        if self.worker is not None:
            if not self.worker.isFinished():
                self.worker.requestInterruption()
                self.worker.wait()

        installed = self.addon.status() != Addon.Status.NOT_INSTALLED
        self.ui.set_installed(installed)
        update_info = UpdateInformation()
        if installed:
            update_info.unchecked = self.addon.status() == Addon.Status.UNCHECKED
            update_info.update_available = self.addon.status() == Addon.Status.UPDATE_AVAILABLE
            update_info.check_in_progress = False  # TODO: Implement the "check in progress" status
            if repo.metadata:
                update_info.branch = get_branch_from_metadata(repo.metadata)
                update_info.version = repo.metadata.version
            elif repo.macro:
                update_info.version = repo.macro.version
            self.ui.set_update_available(update_info)
            self.ui.set_location(os.path.join(self.addon.mod_directory, self.addon.name))
            self.ui.set_location(os.path.join(self.addon.mod_directory, self.addon.name))
            self.ui.set_disabled(self.addon.is_disabled())
        self.ui.allow_running(repo.repo_type == Addon.Kind.MACRO)
        self.ui.allow_disabling(repo.repo_type != Addon.Kind.MACRO)

        if repo.repo_type == Addon.Kind.MACRO:
            self.update_macro_info(repo)

        if repo.status() == Addon.Status.UNCHECKED:
            self.ui.button_bar.check_for_update.show()
            self.ui.button_bar.check_for_update.setText(
                translate("AddonsInstaller", "Check for " "update")
            )
            self.ui.button_bar.check_for_update.setEnabled(True)
            if not self.update_check_thread:
                self.update_check_thread = QtCore.QThread()
            self.check_for_update_worker = CheckSingleUpdateWorker(repo)
            self.check_for_update_worker.moveToThread(self.update_check_thread)
            self.update_check_thread.finished.connect(self.check_for_update_worker.deleteLater)
            self.ui.button_bar.check_for_update.clicked.connect(
                self.check_for_update_worker.do_work
            )
            self.check_for_update_worker.update_status.connect(self.display_repo_status)
            self.update_check_thread.start()
        else:
            self.ui.button_bar.check_for_update.hide()

        flags = WarningFlags()
        flags.required_freecad_version = self.requires_newer_freecad()
        flags.obsolete = repo.obsolete
        flags.python2 = repo.python2
        self.ui.set_warning_flags(flags)
        self.set_change_branch_button_state()

    def requires_newer_freecad(self) -> Optional[Version]:
        """If the current package is not installed, returns the first supported version of
        FreeCAD, if one is set, or None if no information is available (or if the package is
        already installed)."""

        # If it's not installed, check to see if it's for a newer version of FreeCAD
        if self.addon.status() == Addon.Status.NOT_INSTALLED and self.addon.metadata:
            # Only hide if ALL content items require a newer version, otherwise
            # it's possible that this package actually provides versions of itself
            # for newer and older versions

            first_supported_version = get_first_supported_freecad_version(self.addon.metadata)
            if first_supported_version is not None:
                fc_version = Version(from_list=fci.Version())
                if first_supported_version > fc_version:
                    return first_supported_version
        return None

    def set_change_branch_button_state(self):
        """The change branch button is only available for installed Addons that have a .git directory
        and in runs where the git is available."""

        self.ui.button_bar.change_branch.hide()

        pref = fci.ParamGet("User parameter:BaseApp/Preferences/Addons")
        show_switcher = pref.GetBool("ShowBranchSwitcher", False)
        if not show_switcher:
            return

        # Is this repo installed? If not, return.
        if self.addon.status() == Addon.Status.NOT_INSTALLED:
            return

        # Is it a Macro? If so, return:
        if self.addon.repo_type == Addon.Kind.MACRO:
            return

        # Can we actually switch branches? If not, return.
        if not self.git_manager:
            return

        # Is there a .git subdirectory? If not, return.
        basedir = fci.getUserAppDataDir()
        path_to_git = os.path.join(basedir, "Mod", self.addon.name, ".git")
        if not os.path.isdir(path_to_git):
            return

        # If all four above checks passed, then it's possible for us to switch
        # branches, if there are any besides the one we are on: show the button
        self.ui.button_bar.change_branch.show()

    def update_macro_info(self, repo: Addon) -> None:
        if not repo.macro.url:
            # We need to populate the macro information... may as well do it while the user reads
            # the wiki page
            self.worker = GetMacroDetailsWorker(repo)
            self.worker.readme_updated.connect(self.macro_readme_updated)
            self.worker.start()

    def change_branch_clicked(self) -> None:
        """Loads the branch-switching dialog"""
        basedir = fci.getUserAppDataDir()
        path_to_repo = os.path.join(basedir, "Mod", self.addon.name)
        change_branch_dialog = ChangeBranchDialog(path_to_repo, self.ui)
        change_branch_dialog.branch_changed.connect(self.branch_changed)
        change_branch_dialog.exec()

    def enable_clicked(self) -> None:
        """Called by the Enable button, enables this Addon and updates GUI to reflect
        that status."""
        self.addon.enable()
        self.ui.set_disabled(False)
        if self.original_disabled_state:
            self.ui.set_new_disabled_status(False)
            self.original_status = self.addon.status()
            self.addon.set_status(Addon.Status.PENDING_RESTART)
        else:
            self.addon.set_status(self.original_status)
        self.update_status.emit(self.addon)

    def disable_clicked(self) -> None:
        """Called by the Disable button, disables this Addon and updates the GUI to
        reflect that status."""
        self.addon.disable()
        self.ui.set_disabled(True)
        if not self.original_disabled_state:
            self.ui.set_new_disabled_status(True)
            self.original_status = self.addon.status()
            self.addon.set_status(Addon.Status.PENDING_RESTART)
        else:
            self.addon.set_status(self.original_status)
        self.update_status.emit(self.addon)

    def branch_changed(self, name: str) -> None:
        """Displays a dialog confirming the branch changed, and tries to access the
        metadata file from that branch."""
        QtWidgets.QMessageBox.information(
            self.ui,
            translate("AddonsInstaller", "Success"),
            translate(
                "AddonsInstaller",
                "Branch change succeeded, please restart to use the new version.",
            ),
        )
        # See if this branch has a package.xml file:
        basedir = fci.getUserAppDataDir()
        path_to_metadata = os.path.join(basedir, "Mod", self.addon.name, "package.xml")
        if os.path.isfile(path_to_metadata):
            self.addon.load_metadata_file(path_to_metadata)
            self.addon.installed_version = self.addon.metadata.version
        else:
            self.addon.repo_type = Addon.Kind.WORKBENCH
            self.addon.metadata = None
            self.addon.installed_version = None
        self.addon.updated_timestamp = QtCore.QDateTime.currentDateTime().toSecsSinceEpoch()
        self.addon.branch = name
        self.addon.set_status(Addon.Status.PENDING_RESTART)
        self.ui.set_new_branch(name)
        self.update_status.emit(self.addon)

    def display_repo_status(self, addon):
        self.update_status.emit(self.addon)
        self.show_repo(self.addon)

    def macro_readme_updated(self):
        self.show_repo(self.addon)
